# SPDX-FileCopyrightText: 2025 Lucas S
#
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass
from datetime import UTC, datetime, timedelta, tzinfo

from jararaca.messagebus.publisher import IMessage, MessagePublisher


@dataclass
class CollectorDelayedMessageData:
    message: IMessage
    when: datetime
    timezone: tzinfo


class MessagePublisherCollector(MessagePublisher):

    def __init__(self) -> None:
        self.staged_delayed_messages: list[CollectorDelayedMessageData] = []
        self.staged_messages: list[IMessage] = []

    async def publish(self, message: IMessage, topic: str) -> None:
        self.staged_messages.append(message)

    async def delay(self, message: IMessage, seconds: int) -> None:
        self.staged_delayed_messages.append(
            CollectorDelayedMessageData(
                message=message,
                when=datetime.now(UTC) + timedelta(seconds=seconds),
                timezone=UTC,
            )
        )

    async def schedule(
        self, message: IMessage, when: datetime, timezone: tzinfo
    ) -> None:
        self.staged_delayed_messages.append(
            CollectorDelayedMessageData(
                message=message,
                when=when,
                timezone=timezone,
            )
        )

    async def fill(self, publisher: MessagePublisher) -> None:
        for message in self.staged_messages:
            await publisher.publish(message, message.MESSAGE_TOPIC)

        for delayed_message in self.staged_delayed_messages:
            await publisher.schedule(
                delayed_message.message,
                delayed_message.when,
                delayed_message.timezone,
            )

    def has_messages(self) -> bool:
        return bool(self.staged_messages or self.staged_delayed_messages)

    async def flush(self) -> None:
        raise NotImplementedError("I'm just a poor little collector! :(")



# SPDX-FileCopyrightText: 2025 Lucas S
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any, Callable, Protocol, Self, Type, TypeVar

from pydantic import BaseModel
from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase

IDENTIFIABLE_SCHEMA_T = TypeVar("IDENTIFIABLE_SCHEMA_T")


T_BASEMODEL = TypeVar("T_BASEMODEL", bound=BaseModel)


def recursive_get_dict(obj: Any) -> Any:
    if isinstance(obj, list):
        return [recursive_get_dict(v) for v in obj]
    elif hasattr(obj, "__dict__"):
        return {
            k: recursive_get_dict(v) for k, v in obj.__dict__.items() if k[0] != "_"
        }
    else:
        return obj


RESULT_T = TypeVar("RESULT_T", covariant=True)
ENTITY_T_CONTRA = TypeVar("ENTITY_T_CONTRA", bound="BaseEntity", contravariant=True)


class EntityParserType(Protocol[ENTITY_T_CONTRA, RESULT_T]):

    @classmethod
    def parse_entity(cls, model: ENTITY_T_CONTRA) -> "RESULT_T": ...


EntityParserFunc = Callable[[ENTITY_T_CONTRA], RESULT_T]

BASED_BASE_ENTITY_T = TypeVar("BASED_BASE_ENTITY_T", bound="BaseEntity")


class BaseEntity(AsyncAttrs, DeclarativeBase):

    @classmethod
    def from_basemodel(cls, mutation: BaseModel) -> "Self":
        intersection = set(cls.__annotations__.keys()) & set(
            mutation.__class__.model_fields.keys()
        )
        return cls(**{k: getattr(mutation, k) for k in intersection})

    def to_basemodel(self, model: Type[T_BASEMODEL]) -> T_BASEMODEL:
        return model.model_validate(recursive_get_dict(self))

    def parse_entity_with_func(
        self, model_cls: EntityParserFunc["Self", RESULT_T]
    ) -> RESULT_T:
        return model_cls(self)

    def parse_entity_with_type(
        self: BASED_BASE_ENTITY_T,
        model_cls: Type[EntityParserType[BASED_BASE_ENTITY_T, RESULT_T]],
    ) -> RESULT_T:
        return model_cls.parse_entity(self)

    def __rshift__(self, model: EntityParserFunc["Self", RESULT_T]) -> RESULT_T:
        return self.parse_entity_with_func(model)

    def __and__(self, model: Type[EntityParserType["Self", RESULT_T]]) -> RESULT_T:
        return self.parse_entity_with_type(model)



# SPDX-FileCopyrightText: 2025 Lucas S
#
# SPDX-License-Identifier: GPL-3.0-or-later


import inspect
from dataclasses import dataclass
from typing import Any, Awaitable, Callable, Literal, TypeVar, cast, get_args

from jararaca.messagebus.message import INHERITS_MESSAGE_CO, Message, MessageOf
from jararaca.reflect.controller_inspect import (
    ControllerMemberReflect,
    inspect_controller,
)
from jararaca.reflect.decorators import (
    FUNC_OR_TYPE_T,
    GenericStackableDecorator,
    StackableDecorator,
)
from jararaca.reflect.helpers import is_generic_alias
from jararaca.scheduler.decorators import ScheduledAction, ScheduledActionData
from jararaca.utils.env_parse_utils import get_env_float, get_env_int, is_env_truffy
from jararaca.utils.retry import RetryPolicy

AcceptableHandler = (
    Callable[[Any, MessageOf[Any]], Awaitable[None]]
    | Callable[[Any, Any], Awaitable[None]]
)
MessageHandlerT = TypeVar("MessageHandlerT", bound=AcceptableHandler)

DEFAULT_TIMEOUT = get_env_int("JARARACA_MESSAGEBUS_HANDLER_TIMEOUT")
DEFAULT_NACK_ON_EXCEPTION = is_env_truffy("JARARACA_MESSAGEBUS_NACK_ON_EXCEPTION")
DEFAULT_AUTO_ACK = is_env_truffy("JARARACA_MESSAGEBUS_AUTO_ACK")
DEFAULT_NACK_DELAY_ON_EXCEPTION = get_env_float(
    "JARARACA_MESSAGEBUS_NACK_DELAY_ON_EXCEPTION"
)


class MessageHandler(GenericStackableDecorator[AcceptableHandler]):

    def __init__(
        self,
        message: type[INHERITS_MESSAGE_CO],
        *,
        timeout: int | None = DEFAULT_TIMEOUT if DEFAULT_TIMEOUT is not False else None,
        exception_handler: Callable[[BaseException], None] | None = None,
        nack_on_exception: bool = DEFAULT_NACK_ON_EXCEPTION,
        nack_delay_on_exception: float = DEFAULT_NACK_DELAY_ON_EXCEPTION or 5.0,
        auto_ack: bool = DEFAULT_AUTO_ACK,
        name: str | None = None,
        retry_config: RetryPolicy | None = None,
    ) -> None:
        self.message_type = message

        self.timeout = timeout
        self.exception_handler = exception_handler
        self.nack_on_exception = nack_on_exception
        self.nack_delay_on_exception = nack_delay_on_exception

        self.auto_ack = auto_ack
        self.name = name
        self.retry_config = retry_config

    def __call__(self, subject: MessageHandlerT) -> MessageHandlerT:
        return cast(MessageHandlerT, super().__call__(subject))

    def pre_decorated(self, subject: FUNC_OR_TYPE_T) -> None:
        MessageHandler.validate_decorated_fn(subject)

    @staticmethod
    def validate_decorated_fn(
        subject: FUNC_OR_TYPE_T,
    ) -> tuple[Literal["WRAPPED", "DIRECT"], type[Message]]:
        """Validates that the decorated function has the correct signature
        the decorated must follow one of the patterns:

        async def handler(self, message: MessageOf[YourMessageType]) -> None:
            ...

        async def handler(self, message: YourMessageType) -> None:
            ...

        """

        if not inspect.iscoroutinefunction(subject):
            raise RuntimeError(
                "Message handler '%s' must be a coroutine function"
                % (subject.__qualname__)
            )

        signature = inspect.signature(subject)

        parameters = list(signature.parameters.values())

        if len(parameters) != 2:
            raise RuntimeError(
                "Message handler '%s' must have exactly two parameters (self, message)"
                % (subject.__qualname__)
            )

        message_param = parameters[1]

        if message_param.annotation is inspect.Parameter.empty:
            raise RuntimeError(
                "Message handler '%s' must have type annotation for the message parameter"
                % (subject.__qualname__)
            )

        annotation_type = message_param.annotation
        mode: Literal["WRAPPED", "DIRECT"]
        if is_generic_alias(annotation_type):

            message_model_type = get_args(annotation_type)[0]

            mode = "WRAPPED"

        else:
            message_model_type = annotation_type

            mode = "DIRECT"

        if not inspect.isclass(message_model_type) or not issubclass(
            message_model_type, Message
        ):
            raise RuntimeError(
                "Message handler '%s' message parameter must be of type 'MessageOf[YourMessageType]' or 'YourMessageType' where 'YourMessageType' is a subclass of 'Message'"
                % (subject.__qualname__)
            )

        return mode, message_model_type


@dataclass(frozen=True)
class MessageHandlerData:
    message_type: type[Any]
    spec: MessageHandler
    instance_callable: Callable[..., Awaitable[None]]
    controller_member: ControllerMemberReflect


@dataclass(frozen=True)
class ScheduleDispatchData:
    timestamp: float


SCHEDULED_ACTION_DATA_SET = set[ScheduledActionData]

MESSAGE_HANDLER_DATA_SET = set[MessageHandlerData]


class MessageBusController(StackableDecorator):

    def __init__(
        self,
        *,
        inherit_class_decorators: bool = True,
        inherit_methods_decorators: bool = True,
    ) -> None:
        self.messagebus_factory: (
            Callable[[Any], tuple[MESSAGE_HANDLER_DATA_SET, SCHEDULED_ACTION_DATA_SET]]
            | None
        ) = None

        self.inherit_class_decorators = inherit_class_decorators
        self.inherit_methods_decorators = inherit_methods_decorators

    def get_messagebus_factory(
        self,
    ) -> Callable[
        [FUNC_OR_TYPE_T], tuple[MESSAGE_HANDLER_DATA_SET, SCHEDULED_ACTION_DATA_SET]
    ]:
        if self.messagebus_factory is None:
            raise Exception("MessageBus factory is not set")
        return self.messagebus_factory

    def post_decorated(self, subject: FUNC_OR_TYPE_T) -> None:

        def messagebus_factory(
            instance: FUNC_OR_TYPE_T,
        ) -> tuple[MESSAGE_HANDLER_DATA_SET, SCHEDULED_ACTION_DATA_SET]:
            handlers: MESSAGE_HANDLER_DATA_SET = set()

            schedulers: SCHEDULED_ACTION_DATA_SET = set()

            assert inspect.isclass(
                subject
            ), "MessageBusController can only be applied to classes"

            _, members = inspect_controller(subject)

            for name, member in members.items():
                message_handler_decoration = MessageHandler.get_last(
                    member.member_function
                )
                scheduled_action_decoration = ScheduledAction.get_last(
                    member.member_function
                )

                if message_handler_decoration is not None:

                    if not inspect.iscoroutinefunction(member.member_function):
                        raise Exception(
                            "Message incoming handler '%s' from '%s.%s' must be a coroutine function"
                            % (name, subject.__module__, subject.__qualname__)
                        )

                    handlers.add(
                        MessageHandlerData(
                            message_type=message_handler_decoration.message_type,
                            spec=message_handler_decoration,
                            instance_callable=getattr(instance, name),
                            controller_member=member,
                        )
                    )
                elif scheduled_action_decoration is not None:
                    if not inspect.iscoroutinefunction(member.member_function):
                        raise Exception(
                            "Scheduled action handler '%s' from '%s.%s' must be a coroutine function"
                            % (name, subject.__module__, subject.__qualname__)
                        )
                    instance_callable = getattr(instance, name)

                    schedulers.add(
                        ScheduledActionData(
                            controller_member=member,
                            spec=scheduled_action_decoration,
                            callable=instance_callable,
                        )
                    )

            return handlers, schedulers

        self.messagebus_factory = messagebus_factory



# SPDX-FileCopyrightText: 2025 Lucas S
#
# SPDX-License-Identifier: GPL-3.0-or-later

import asyncio
import logging
from dataclasses import dataclass
from typing import Any

from redis.asyncio import Redis

from jararaca.presentation.websocket.websocket_interceptor import (
    BroadcastFunc,
    SendFunc,
    WebSocketConnectionBackend,
)

logger = logging.getLogger(__name__)


@dataclass
class BroadcastMessage:
    message: bytes

    def encode(self) -> bytes:
        return self.message

    @staticmethod
    def decode(data: bytes) -> "BroadcastMessage":
        return BroadcastMessage(message=data)

    @staticmethod
    def from_message(message: bytes) -> "BroadcastMessage":
        return BroadcastMessage(message=message)


@dataclass
class SendToRoomsMessage:
    rooms: list[str]
    message: bytes

    def encode(self) -> bytes:
        return b"\x00".join([b",".join([a.encode() for a in self.rooms]), self.message])

    @staticmethod
    def decode(data: bytes) -> "SendToRoomsMessage":
        rooms, message = data.split(b"\x00")
        return SendToRoomsMessage(
            rooms=[a.decode() for a in rooms.split(b",")], message=message
        )

    @staticmethod
    def from_message(rooms: list[str], message: bytes) -> "SendToRoomsMessage":
        return SendToRoomsMessage(rooms=rooms, message=message)


class RedisWebSocketConnectionBackend(WebSocketConnectionBackend):
    def __init__(
        self,
        conn: "Redis[bytes]",
        broadcast_pubsub_channel: str,
        send_pubsub_channel: str,
        consume_broadcast_timeout: int = 1,
        consume_send_timeout: int = 1,
        retry_delay: float = 5.0,
        max_concurrent_tasks: int = 1000,
    ) -> None:

        self.redis = conn
        self.broadcast_pubsub_channel = broadcast_pubsub_channel
        self.send_pubsub_channel = send_pubsub_channel

        self.lock = asyncio.Lock()
        self.tasks: set[asyncio.Task[Any]] = set()
        self.max_concurrent_tasks = max_concurrent_tasks
        self.task_semaphore = asyncio.Semaphore(max_concurrent_tasks)

        self.consume_broadcast_timeout = consume_broadcast_timeout
        self.consume_send_timeout = consume_send_timeout
        self.retry_delay = retry_delay
        self.__shutdown_event: asyncio.Event | None = None

        self.__send_func: SendFunc | None = None
        self.__broadcast_func: BroadcastFunc | None = None

    @property
    def shutdown_event(self) -> asyncio.Event:
        if self.__shutdown_event is None:
            raise RuntimeError(
                "Shutdown event is not set. Please configure the backend before using it."
            )
        return self.__shutdown_event

    @property
    def send_func(self) -> SendFunc:
        if self.__send_func is None:
            raise RuntimeError(
                "Send function is not set. Please configure the backend before using it."
            )
        return self.__send_func

    @property
    def broadcast_func(self) -> BroadcastFunc:
        if self.__broadcast_func is None:
            raise RuntimeError(
                "Broadcast function is not set. Please configure the backend before using it."
            )
        return self.__broadcast_func

    async def broadcast(self, message: bytes) -> None:
        try:
            await self.redis.publish(
                self.broadcast_pubsub_channel,
                BroadcastMessage.from_message(message).encode(),
            )
        except Exception as e:
            logger.error(
                "Failed to publish broadcast message to Redis: %s", e, exc_info=True
            )
            raise

    async def send(self, rooms: list[str], message: bytes) -> None:
        try:
            await self.redis.publish(
                self.send_pubsub_channel,
                SendToRoomsMessage.from_message(rooms, message).encode(),
            )
        except Exception as e:
            logger.error(
                "Failed to publish send message to Redis: %s", e, exc_info=True
            )
            raise

    def configure(
        self, broadcast: BroadcastFunc, send: SendFunc, shutdown_event: asyncio.Event
    ) -> None:
        if self.__shutdown_event is not None:
            raise RuntimeError("Backend is already configured.")
        self.__shutdown_event = shutdown_event
        self.__send_func = send
        self.__broadcast_func = broadcast
        self.setup_send_consumer()
        self.setup_broadcast_consumer()

    def setup_send_consumer(self) -> None:

        send_task = asyncio.get_event_loop().create_task(
            self.consume_send(self.send_func, self.shutdown_event)
        )

        # Use lock when modifying tasks set to prevent race conditions
        async def add_task() -> None:
            async with self.lock:
                self.tasks.add(send_task)

        asyncio.get_event_loop().create_task(add_task())
        send_task.add_done_callback(self.handle_send_task_done)

    def setup_broadcast_consumer(self) -> None:

        broadcast_task = asyncio.get_event_loop().create_task(
            self.consume_broadcast(self.broadcast_func, self.shutdown_event)
        )

        # Use lock when modifying tasks set to prevent race conditions
        async def add_task() -> None:
            async with self.lock:
                self.tasks.add(broadcast_task)

        asyncio.get_event_loop().create_task(add_task())

        broadcast_task.add_done_callback(self.handle_broadcast_task_done)

    def handle_broadcast_task_done(self, task: asyncio.Task[Any]) -> None:
        # Remove task from set safely with lock
        async def remove_task() -> None:
            async with self.lock:
                self.tasks.discard(task)

        asyncio.get_event_loop().create_task(remove_task())

        if task.cancelled():
            logger.warning("Broadcast task was cancelled.")
        elif task.exception() is not None:
            logger.exception(
                "Broadcast task raised an exception:", exc_info=task.exception()
            )
        else:
            logger.warning("Broadcast task somehow completed successfully.")

        if not self.shutdown_event.is_set():
            logger.warning(
                "Broadcast task completed, but shutdown event is not set. This is unexpected."
            )
            # Add delay before retrying to avoid excessive CPU usage
            asyncio.get_event_loop().create_task(
                self._retry_broadcast_consumer_with_delay()
            )

    def handle_send_task_done(self, task: asyncio.Task[Any]) -> None:
        # Remove task from set safely with lock
        async def remove_task() -> None:
            async with self.lock:
                self.tasks.discard(task)

        asyncio.get_event_loop().create_task(remove_task())

        if task.cancelled():
            logger.warning("Send task was cancelled.")
        elif task.exception() is not None:
            logger.exception(
                "Send task raised an exception:", exc_info=task.exception()
            )
        else:
            logger.warning("Send task somehow completed successfully.")

        if not self.shutdown_event.is_set():
            logger.warning(
                "Send task completed, but shutdown event is not set. This is unexpected."
            )
            # Add delay before retrying to avoid excessive CPU usage
            asyncio.get_event_loop().create_task(self._retry_send_consumer_with_delay())

    async def _retry_broadcast_consumer_with_delay(self) -> None:
        """Retry setting up broadcast consumer after a delay to avoid excessive CPU usage."""
        logger.warning(
            "Waiting %s seconds before retrying broadcast consumer...", self.retry_delay
        )
        await asyncio.sleep(self.retry_delay)

        if not self.shutdown_event.is_set():
            logger.warning("Retrying broadcast consumer setup...")
            self.setup_broadcast_consumer()

    async def _retry_send_consumer_with_delay(self) -> None:
        """Retry setting up send consumer after a delay to avoid excessive CPU usage."""
        logger.warning(
            "Waiting %s seconds before retrying send consumer...", self.retry_delay
        )
        await asyncio.sleep(self.retry_delay)

        if not self.shutdown_event.is_set():
            logger.warning("Retrying send consumer setup...")
            self.setup_send_consumer()

    async def consume_broadcast(
        self, broadcast: BroadcastFunc, shutdown_event: asyncio.Event
    ) -> None:
        logger.debug("Starting broadcast consumer...")
        try:
            # Validate Redis connection before starting
            try:
                await self.redis.ping()
                logger.debug("Redis connection validated for broadcast consumer")
            except Exception as e:
                logger.error("Redis connection validation failed: %s", e, exc_info=True)
                raise

            async with self.redis.pubsub() as pubsub:
                await pubsub.subscribe(self.broadcast_pubsub_channel)
                logger.debug(
                    "Subscribed to broadcast channel: %s", self.broadcast_pubsub_channel
                )

                while not shutdown_event.is_set():
                    message: dict[str, Any] | None = await pubsub.get_message(
                        ignore_subscribe_messages=True,
                        timeout=self.consume_broadcast_timeout,
                    )

                    if message is None:
                        continue

                    broadcast_message = BroadcastMessage.decode(message["data"])

                    # Use semaphore for backpressure control
                    acquired = False
                    try:
                        await self.task_semaphore.acquire()
                        acquired = True

                        async def broadcast_with_cleanup(msg: bytes) -> None:
                            try:
                                await broadcast(message=msg)
                            finally:
                                self.task_semaphore.release()

                        async with self.lock:
                            task = asyncio.get_event_loop().create_task(
                                broadcast_with_cleanup(broadcast_message.message)
                            )

                            self.tasks.add(task)

                            task.add_done_callback(self.tasks.discard)
                    except Exception as e:
                        # Release semaphore if we acquired it but failed to create task
                        if acquired:
                            self.task_semaphore.release()
                        logger.error(
                            "Error processing broadcast message: %s", e, exc_info=True
                        )
                        # Continue processing other messages
                        continue
        except Exception as e:
            logger.error(
                "Fatal error in broadcast consumer, will retry: %s", e, exc_info=True
            )
            raise

    async def consume_send(self, send: SendFunc, shutdown_event: asyncio.Event) -> None:
        logger.debug("Starting send consumer...")
        try:
            # Validate Redis connection before starting
            try:
                await self.redis.ping()
                logger.debug("Redis connection validated for send consumer")
            except Exception as e:
                logger.error("Redis connection validation failed: %s", e, exc_info=True)
                raise

            async with self.redis.pubsub() as pubsub:
                await pubsub.subscribe(self.send_pubsub_channel)
                logger.debug("Subscribed to send channel: %s", self.send_pubsub_channel)

                while not shutdown_event.is_set():
                    message: dict[str, Any] | None = await pubsub.get_message(
                        ignore_subscribe_messages=True,
                        timeout=self.consume_send_timeout,
                    )

                    if message is None:
                        continue

                    send_message = SendToRoomsMessage.decode(message["data"])

                    # Use semaphore for backpressure control
                    acquired = False
                    try:
                        await self.task_semaphore.acquire()
                        acquired = True

                        async def send_with_cleanup(
                            rooms: list[str], msg: bytes
                        ) -> None:
                            try:
                                await send(rooms, msg)
                            finally:
                                self.task_semaphore.release()

                        async with self.lock:

                            task = asyncio.get_event_loop().create_task(
                                send_with_cleanup(
                                    send_message.rooms, send_message.message
                                )
                            )

                            self.tasks.add(task)

                            task.add_done_callback(self.tasks.discard)
                    except Exception as e:
                        # Release semaphore if we acquired it but failed to create task
                        if acquired:
                            self.task_semaphore.release()
                        logger.error(
                            "Error processing send message: %s", e, exc_info=True
                        )
                        # Continue processing other messages
                        continue
        except Exception as e:
            logger.error(
                "Fatal error in send consumer, will retry: %s", e, exc_info=True
            )
            raise

    async def shutdown(self) -> None:
        async with self.lock:

            await asyncio.gather(*self.tasks, return_exceptions=True)

            await self.redis.close()



# SPDX-FileCopyrightText: 2025 Lucas S
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any, Type, TypeVar

from pydantic import BaseModel

from jararaca.core.providers import Token
from jararaca.reflect.decorators import StackableDecorator

DECORATED_CLASS = TypeVar("DECORATED_CLASS", bound=Any)


class RequiresConfig(StackableDecorator):

    def __init__(self, token: Token[Any], config: Type[BaseModel]):
        self.config = config
        self.token = token



# SPDX-FileCopyrightText: 2025 Lucas S
#
# SPDX-License-Identifier: GPL-3.0-or-later

import inspect
from contextlib import contextmanager, suppress
from contextvars import ContextVar
from functools import wraps
from typing import (
    Any,
    AsyncContextManager,
    Awaitable,
    Callable,
    ContextManager,
    Generator,
    Literal,
    Mapping,
    Protocol,
    Sequence,
    TypeVar,
    Union,
)

from jararaca.microservice import AppTransactionContext

F = TypeVar("F", bound=Callable[..., Awaitable[Any]])

AttributeValue = Union[
    str,
    bool,
    int,
    float,
    Sequence[str],
    Sequence[bool],
    Sequence[int],
    Sequence[float],
]

AttributeMap = Mapping[str, AttributeValue]


class TracingSpan(Protocol): ...


class TracingSpanContext(Protocol): ...


class TracingContextProvider(Protocol):

    def start_span_context(
        self, trace_name: str, context_attributes: AttributeMap | None
    ) -> ContextManager[Any]: ...

    def add_event(
        self,
        event_name: str,
        event_attributes: AttributeMap | None = None,
    ) -> None: ...

    def set_span_status(self, status_code: Literal["OK", "ERROR", "UNSET"]) -> None: ...

    def record_exception(
        self,
        exception: Exception,
        attributes: AttributeMap | None = None,
        escaped: bool = False,
    ) -> None: ...

    def set_span_attribute(
        self,
        key: str,
        value: AttributeValue,
    ) -> None: ...

    def update_span_name(self, new_name: str) -> None: ...

    def add_link(self, span_context: TracingSpanContext) -> None: ...

    def get_current_span(self) -> TracingSpan | None: ...
    def get_current_span_context(self) -> TracingSpanContext | None: ...


class TracingContextProviderFactory(Protocol):

    def root_setup(
        self, app_context: AppTransactionContext
    ) -> AsyncContextManager[None]: ...

    def provide_provider(
        self, app_context: AppTransactionContext
    ) -> TracingContextProvider: ...


tracing_ctx_provider_ctxv = ContextVar[TracingContextProvider]("tracing_ctx_provider")


@contextmanager
def provide_tracing_ctx_provider(
    ctx_provider: TracingContextProvider,
) -> Generator[None, None, None]:

    token = tracing_ctx_provider_ctxv.set(ctx_provider)
    try:
        yield
    finally:
        with suppress(ValueError):
            tracing_ctx_provider_ctxv.reset(token)


def get_tracing_ctx_provider() -> TracingContextProvider | None:
    return tracing_ctx_provider_ctxv.get(None)


def default_trace_mapper(*args: Any, **kwargs: Any) -> dict[str, str]:
    return {
        "args": str(args),
        "kwargs": str(kwargs),
    }


class TracedFunc:

    def __init__(
        self,
        trace_name: str,
        trace_mapper: Callable[..., dict[str, str]] = default_trace_mapper,
    ):
        self.trace_name = trace_name
        self.trace_mapper = trace_mapper

    def __call__(
        self,
        decorated: F,
    ) -> F:

        @wraps(decorated)
        async def wrapper(
            *args: Any,
            **kwargs: Any,
        ) -> Any:

            if ctx_provider := get_tracing_ctx_provider():
                with ctx_provider.start_span_context(
                    self.trace_name,
                    self.trace_mapper(*args, **kwargs),
                ):
                    return await decorated(*args, **kwargs)

            return await decorated(*args, **kwargs)

        return wrapper  # type: ignore[return-value]


C = TypeVar("C", bound=type)


class TracedClass:
    """
    Class decorator that automatically applies tracing to all async methods in a class.

    Usage:
        @TracedClass()
        class MyService:
            async def method1(self) -> str:
                return "hello"

            async def method2(self, x: int) -> int:
                return x * 2

            def sync_method(self) -> str:  # Not traced
                return "sync"
    """

    def __init__(
        self,
        trace_name_prefix: str | None = None,
        trace_mapper: Callable[..., dict[str, str]] = default_trace_mapper,
        include_private: bool = False,
        exclude_methods: set[str] | None = None,
    ):
        """
        Initialize the TracedClass decorator.

        Args:
            trace_name_prefix: Prefix for trace names. If None, uses class name.
            trace_mapper: Function to map method arguments to trace attributes.
            include_private: Whether to trace private methods (starting with _).
            exclude_methods: Set of method names to exclude from tracing.
        """
        self.trace_name_prefix = trace_name_prefix
        self.trace_mapper = trace_mapper
        self.include_private = include_private
        self.exclude_methods = exclude_methods or set()

    def __call__(self, cls: C) -> C:
        """Apply tracing to all async methods in the class."""

        # Use class name as prefix if not provided
        trace_prefix = self.trace_name_prefix or cls.__name__

        # Get all methods in the class
        for name, method in inspect.getmembers_static(
            cls, predicate=inspect.isfunction
        ):
            # Skip if method should be excluded
            if name in self.exclude_methods:
                continue

            # Skip private methods unless explicitly included
            if name.startswith("_") and not self.include_private:
                continue

            # Only trace async methods
            if inspect.iscoroutinefunction(method):
                trace_name = f"{trace_prefix}.{name}"
                traced_method = TracedFunc(trace_name, self.trace_mapper)(method)
                setattr(cls, name, traced_method)

        return cls


def traced_class(
    trace_name_prefix: str | None = None,
    trace_mapper: Callable[..., dict[str, str]] = default_trace_mapper,
    include_private: bool = False,
    exclude_methods: set[str] | None = None,
) -> Callable[[C], C]:
    """
    Functional interface for TracedClass decorator.

    Usage:
        @traced_class(trace_name_prefix="MyService")
        class MyService:
            async def method1(self) -> str:
                return "hello"
    """
    return TracedClass(
        trace_name_prefix=trace_name_prefix,
        trace_mapper=trace_mapper,
        include_private=include_private,
        exclude_methods=exclude_methods,
    )


__all__ = [
    "TracingContextProvider",
    "TracingContextProviderFactory",
    "provide_tracing_ctx_provider",
    "get_tracing_ctx_provider",
    "default_trace_mapper",
    "TracedFunc",
    "TracedClass",
    "traced_class",
]



# SPDX-FileCopyrightText: 2025 Lucas S
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
from typing import Literal, Optional, TypeVar, overload


def is_env_truffy(var_name: str) -> bool:
    value = os.getenv(var_name, "").lower()
    return value in ("1", "true", "yes", "on")


DF_BOOL_T = TypeVar("DF_BOOL_T", bound="bool")


@overload
def get_env_bool(
    var_name: str, default: DF_BOOL_T
) -> DF_BOOL_T | bool | Literal["invalid"]: ...


@overload
def get_env_bool(
    var_name: str, default: None = None
) -> bool | None | Literal["invalid"]: ...


def get_env_bool(
    var_name: str, default: DF_BOOL_T | None = None
) -> DF_BOOL_T | bool | Literal["invalid"] | None:
    value = os.getenv(var_name)
    if value is None:
        return default
    value_lower = value.lower()
    if value_lower in ("1", "true", "yes", "on"):
        return True
    elif value_lower in ("0", "false", "no", "off"):
        return False
    else:
        return "invalid"


DF_INT_T = TypeVar("DF_INT_T", bound="int | None | Literal[False]")


@overload
def get_env_int(var_name: str, default: None = None) -> int | None | Literal[False]: ...


@overload
def get_env_int(
    var_name: str, default: DF_INT_T
) -> DF_INT_T | int | Literal[False]: ...


def get_env_int(
    var_name: str, default: DF_INT_T = False  # type: ignore[assignment]
) -> DF_INT_T | int | Literal[False]:
    value = os.getenv(var_name)
    if value is None:
        return default
    try:
        return int(value)
    except ValueError:
        return False


DF_FLOAT_T = TypeVar("DF_FLOAT_T", bound="float | None | Literal[False]")


@overload
def get_env_float(
    var_name: str, default: None = None
) -> float | None | Literal[False]: ...


@overload
def get_env_float(
    var_name: str, default: DF_FLOAT_T
) -> DF_FLOAT_T | float | Literal[False]: ...


def get_env_float(
    var_name: str, default: DF_FLOAT_T = False  # type: ignore[assignment]
) -> DF_FLOAT_T | float | Literal[False]:
    value = os.getenv(var_name)
    if value is None:
        return default
    try:
        return float(value)
    except ValueError:
        return False


DF_STR_T = TypeVar("DF_STR_T", bound="Optional[str]")


@overload
def get_env_str(var_name: str, default: None = None) -> str | None: ...


@overload
def get_env_str(var_name: str, default: DF_STR_T) -> DF_STR_T | str: ...


def get_env_str(var_name: str, default: DF_STR_T = None) -> DF_STR_T | str:  # type: ignore[assignment]
    value = os.getenv(var_name)
    if value is None:
        return default
    return value


def get_env_list(var_name: str, separator: str = ",") -> list[str]:
    value = os.getenv(var_name, "")
    if not value:
        return []
    return [item.strip() for item in value.split(separator) if item.strip()]


def get_env_dict(
    var_name: str, item_separator: str = ",", key_value_separator: str = "="
) -> dict[str, str]:
    value = os.getenv(var_name, "")
    result: dict[str, str] = {}
    if not value:
        return result
    items = [item.strip() for item in value.split(item_separator) if item.strip()]
    for item in items:
        if key_value_separator in item:
            key, val = item.split(key_value_separator, 1)
            result[key.strip()] = val.strip()
    return result



# SPDX-FileCopyrightText: 2025 Lucas S
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

from aio_pika.abc import AbstractChannel, AbstractExchange, AbstractQueue
from aio_pika.exceptions import AMQPError, ChannelClosed, ChannelNotFoundEntity

logger = logging.getLogger(__name__)


class RabbitmqUtils:

    DEAD_LETTER_EXCHANGE = "dlx"
    DEAD_LETTER_QUEUE = "dlq"

    # Note: get_worker_v1_queue method is already defined above

    DEAD_LETTER_EXCHANGE = "dlx"
    DEAD_LETTER_QUEUE = "dlq"

    @classmethod
    async def get_dl_exchange(cls, channel: AbstractChannel) -> AbstractExchange:
        """
        Get the Dead Letter Exchange (DLX) for the given channel.
        """
        try:
            return await channel.get_exchange(
                cls.DEAD_LETTER_EXCHANGE,
            )
        except ChannelNotFoundEntity as e:
            logger.error(
                "Dead Letter Exchange '%s' does not exist. "
                "Please use the declare command to create it first. Error: %s",
                cls.DEAD_LETTER_EXCHANGE,
                e,
            )
            raise
        except ChannelClosed as e:
            logger.error(
                "Channel closed while getting Dead Letter Exchange '%s'. " "Error: %s",
                cls.DEAD_LETTER_EXCHANGE,
                e,
            )
            raise
        except AMQPError as e:
            logger.error(
                "AMQP error while getting Dead Letter Exchange '%s'. " "Error: %s",
                cls.DEAD_LETTER_EXCHANGE,
                e,
            )
            raise

    @classmethod
    async def declare_dl_exchange(
        cls, channel: AbstractChannel, passive: bool
    ) -> AbstractExchange:
        """
        Declare a Dead Letter Exchange (DLX) for the given channel.
        """

        return await channel.declare_exchange(
            cls.DEAD_LETTER_EXCHANGE,
            passive=passive,
            type="direct",
            durable=True,
            auto_delete=False,
        )

    @classmethod
    async def get_dl_queue(cls, channel: AbstractChannel) -> AbstractQueue:
        """
        Get the Dead Letter Queue (DLQ) for the given channel.
        """
        try:
            return await channel.get_queue(
                cls.DEAD_LETTER_QUEUE,
            )
        except ChannelNotFoundEntity as e:
            logger.error(
                "Dead Letter Queue '%s' does not exist. "
                "Please use the declare command to create it first. Error: %s",
                cls.DEAD_LETTER_QUEUE,
                e,
            )
            raise
        except ChannelClosed as e:
            logger.error(
                "Channel closed while getting Dead Letter Queue '%s'. " "Error: %s",
                cls.DEAD_LETTER_QUEUE,
                e,
            )
            raise
        except AMQPError as e:
            logger.error(
                "AMQP error while getting Dead Letter Queue '%s'. " "Error: %s",
                cls.DEAD_LETTER_QUEUE,
                e,
            )
            raise

    @classmethod
    async def declare_dl_queue(
        cls, channel: AbstractChannel, passive: bool
    ) -> AbstractQueue:
        """
        Declare a Dead Letter Queue (DLQ) for the given queue.
        """

        return await channel.declare_queue(
            cls.DEAD_LETTER_QUEUE,
            durable=True,
            passive=passive,
            arguments={
                "x-dead-letter-exchange": "",
                "x-dead-letter-routing-key": cls.DEAD_LETTER_EXCHANGE,
            },
        )

    @classmethod
    async def get_dl_kit(
        cls,
        channel: AbstractChannel,
    ) -> tuple[AbstractExchange, AbstractQueue]:
        """
        Get the Dead Letter Exchange and Queue (DLX and DLQ) for the given channel.
        """
        try:
            dlx = await cls.get_dl_exchange(channel)
            dlq = await cls.get_dl_queue(channel)
            return dlx, dlq
        except ChannelNotFoundEntity as e:
            logger.error(
                "Dead Letter infrastructure does not exist completely. "
                "Please use the declare command to create it first. Error: %s",
                e,
            )
            raise
        except ChannelClosed as e:
            logger.error(
                "Channel closed while getting Dead Letter infrastructure. " "Error: %s",
                e,
            )
            raise
        except AMQPError as e:
            logger.error(
                "AMQP error while getting Dead Letter infrastructure. " "Error: %s", e
            )
            raise

    @classmethod
    async def declare_dl_kit(
        cls,
        channel: AbstractChannel,
        passive: bool = False,
    ) -> tuple[AbstractExchange, AbstractQueue]:
        """
        Declare a Dead Letter Exchange and Queue (DLX and DLQ) for the given channel.
        """
        dlx = await cls.declare_dl_exchange(channel, passive=passive)
        dlq = await cls.declare_dl_queue(channel, passive=passive)
        await dlq.bind(dlx, routing_key=cls.DEAD_LETTER_EXCHANGE)
        return dlx, dlq

    @classmethod
    async def get_main_exchange(
        cls, channel: AbstractChannel, exchange_name: str
    ) -> AbstractExchange:
        """
        Get the main exchange for the given channel.
        """
        try:
            return await channel.get_exchange(exchange_name)
        except ChannelNotFoundEntity as e:
            logger.error(
                "Exchange '%s' does not exist. "
                "Please use the declare command to create it first. Error: %s",
                exchange_name,
                e,
            )
            raise
        except ChannelClosed as e:
            logger.error(
                "Channel closed while getting exchange '%s'. " "Error: %s",
                exchange_name,
                e,
            )
            raise
        except AMQPError as e:
            logger.error(
                "AMQP error while getting exchange '%s'. " "Error: %s", exchange_name, e
            )
            raise

    @classmethod
    async def declare_main_exchange(
        cls, channel: AbstractChannel, exchange_name: str, passive: bool
    ) -> AbstractExchange:
        """
        Declare a main exchange for the given channel.
        """

        return await channel.declare_exchange(
            exchange_name,
            passive=passive,
            type="topic",
            durable=True,
            auto_delete=False,
        )

    @classmethod
    async def get_queue(
        cls,
        channel: AbstractChannel,
        queue_name: str,
    ) -> AbstractQueue:
        """
        Get a queue with the given name.
        """
        try:
            return await channel.get_queue(queue_name)
        except ChannelNotFoundEntity as e:
            logger.error(
                "Queue '%s' does not exist. "
                "Please use the declare command to create it first. Error: %s",
                queue_name,
                e,
            )
            raise
        except ChannelClosed as e:
            logger.error(
                "Channel closed while getting queue '%s'. " "Error: %s", queue_name, e
            )
            raise
        except AMQPError as e:
            logger.error(
                "AMQP error while getting queue '%s'. " "Error: %s", queue_name, e
            )
            raise

    @classmethod
    async def declare_worker_queue(
        cls,
        channel: AbstractChannel,
        queue_name: str,
        passive: bool = False,
    ) -> AbstractQueue:
        """
        Declare a queue with the given name and properties.
        """

        return await channel.declare_queue(
            queue_name,
            passive=passive,
            durable=True,
            arguments={
                "x-dead-letter-exchange": cls.DEAD_LETTER_EXCHANGE,
                "x-dead-letter-routing-key": cls.DEAD_LETTER_EXCHANGE,
            },
        )

    @classmethod
    async def get_scheduled_action_queue(
        cls,
        channel: AbstractChannel,
        queue_name: str,
    ) -> AbstractQueue:
        """
        Get a scheduled action queue.
        """
        try:
            return await channel.get_queue(queue_name)
        except ChannelNotFoundEntity as e:
            logger.error(
                "Scheduler queue '%s' does not exist. "
                "Please use the declare command to create it first. Error: %s",
                queue_name,
                e,
            )
            raise
        except ChannelClosed as e:
            logger.error(
                "Channel closed while getting scheduler queue '%s'. " "Error: %s",
                queue_name,
                e,
            )
            raise
        except AMQPError as e:
            logger.error(
                "AMQP error while getting scheduler queue '%s'. " "Error: %s",
                queue_name,
                e,
            )
            raise

    @classmethod
    async def declare_scheduled_action_queue(
        cls,
        channel: AbstractChannel,
        queue_name: str,
        passive: bool = False,
    ) -> AbstractQueue:
        """
        Declare a scheduled action queue with simple durable configuration.
        The queue has a max length of 1 to ensure only one scheduled task
        is processed at a time.
        """
        return await channel.declare_queue(
            name=queue_name,
            durable=True,
            passive=passive,
            arguments={
                "x-max-length": 1,
            },
        )

    @classmethod
    async def delete_exchange(
        cls,
        channel: AbstractChannel,
        exchange_name: str,
        if_unused: bool = False,
    ) -> None:
        """
        Delete an exchange.
        """
        try:
            await channel.exchange_delete(
                exchange_name=exchange_name,
                if_unused=if_unused,
            )
        except ChannelNotFoundEntity:
            # Exchange might not exist, which is fine
            logger.debug(
                "Exchange '%s' does not exist, nothing to delete.", exchange_name
            )
        except ChannelClosed as e:
            logger.warning(
                "Channel closed while deleting exchange '%s': %s", exchange_name, e
            )
        except AMQPError as e:
            logger.warning(
                "AMQP error while deleting exchange '%s': %s", exchange_name, e
            )

    @classmethod
    async def delete_queue(
        cls,
        channel: AbstractChannel,
        queue_name: str,
        if_unused: bool = False,
        if_empty: bool = False,
    ) -> None:
        """
        Delete a queue.
        """
        try:
            await channel.queue_delete(
                queue_name=queue_name,
                if_unused=if_unused,
                if_empty=if_empty,
            )
        except ChannelNotFoundEntity:
            # Queue might not exist, which is fine
            logger.debug("Queue '%s' does not exist, nothing to delete.", queue_name)
        except ChannelClosed as e:
            logger.warning(
                "Channel closed while deleting queue '%s': %s", queue_name, e
            )
        except AMQPError as e:
            logger.warning("AMQP error while deleting queue '%s': %s", queue_name, e)

    @classmethod
    async def get_dl_queue_message_count(cls, channel: AbstractChannel) -> int:
        """
        Get the message count in the Dead Letter Queue.
        Returns 0 if the queue doesn't exist.
        """
        try:
            await channel.get_queue(cls.DEAD_LETTER_QUEUE)
            # The declaration property contains the message count
            queue_info = await channel.declare_queue(
                cls.DEAD_LETTER_QUEUE, passive=True
            )
            return queue_info.declaration_result.message_count or 0
        except ChannelNotFoundEntity:
            logger.debug("Dead Letter Queue does not exist.")
            return 0
        except ChannelClosed as e:
            logger.error("Channel closed while getting DLQ message count: %s", e)
            raise
        except AMQPError as e:
            logger.error("AMQP error while getting DLQ message count: %s", e)
            raise

    @classmethod
    async def purge_dl_queue(cls, channel: AbstractChannel) -> int:
        """
        Purge all messages from the Dead Letter Queue.
        Returns the number of messages purged.
        """
        try:
            queue = await channel.get_queue(cls.DEAD_LETTER_QUEUE)
            result = await queue.purge()
            return result.message_count or 0
        except ChannelNotFoundEntity as e:
            logger.error(
                "Dead Letter Queue '%s' does not exist. Error: %s",
                cls.DEAD_LETTER_QUEUE,
                e,
            )
            raise
        except ChannelClosed as e:
            logger.error("Channel closed while purging Dead Letter Queue: %s", e)
            raise
        except AMQPError as e:
            logger.error("AMQP error while purging Dead Letter Queue: %s", e)
            raise



# SPDX-FileCopyrightText: 2025 Lucas S
#
# SPDX-License-Identifier: GPL-3.0-or-later

import asyncio
import inspect
import logging
import random
import signal
import time
import uuid
from abc import ABC
from contextlib import asynccontextmanager, suppress
from dataclasses import dataclass, field
from datetime import UTC, datetime
from typing import Any, AsyncContextManager, AsyncGenerator, Awaitable, Optional, Type
from urllib.parse import parse_qs, urlparse

import aio_pika
import aio_pika.abc
import uvloop
from aio_pika.exceptions import (
    AMQPChannelError,
    AMQPConnectionError,
    AMQPError,
    ChannelClosed,
    ChannelNotFoundEntity,
    ConnectionClosed,
)
from pydantic import ValidationError

from jararaca.broker_backend import MessageBrokerBackend
from jararaca.broker_backend.mapper import get_message_broker_backend_from_url
from jararaca.core.uow import UnitOfWorkContextProvider
from jararaca.di import Container
from jararaca.lifecycle import AppLifecycle
from jararaca.messagebus.bus_message_controller import (
    BusMessageController,
    provide_bus_message_controller,
)
from jararaca.messagebus.decorators import (
    MESSAGE_HANDLER_DATA_SET,
    SCHEDULED_ACTION_DATA_SET,
    MessageBusController,
    MessageHandler,
    MessageHandlerData,
    ScheduleDispatchData,
)
from jararaca.messagebus.implicit_headers import provide_implicit_headers
from jararaca.messagebus.message import Message, MessageOf
from jararaca.microservice import (
    AppTransactionContext,
    MessageBusTransactionData,
    Microservice,
    SchedulerTransactionData,
    ShutdownState,
    provide_shutdown_state,
    providing_app_type,
)
from jararaca.observability.hooks import record_exception, set_span_status
from jararaca.scheduler.decorators import ScheduledActionData
from jararaca.utils.rabbitmq_utils import RabbitmqUtils
from jararaca.utils.retry import RetryPolicy, retry_with_backoff

logger = logging.getLogger(__name__)


@dataclass
class AioPikaWorkerConfig:
    url: str
    exchange: str
    prefetch_count: int
    connection_retry_config: RetryPolicy = field(
        default_factory=lambda: RetryPolicy(
            max_retries=15,
            initial_delay=1.0,
            max_delay=60.0,
            backoff_factor=2.0,
        )
    )
    consumer_retry_policy: RetryPolicy = field(
        default_factory=lambda: RetryPolicy(
            max_retries=15,
            initial_delay=0.5,
            max_delay=40.0,
            backoff_factor=2.0,
        )
    )
    # Connection health monitoring settings
    connection_heartbeat_interval: float = 30.0  # seconds
    connection_health_check_interval: float = 10.0  # seconds


class AioPikaMessage(MessageOf[Message]):

    def __init__(
        self,
        aio_pika_message: aio_pika.abc.AbstractIncomingMessage,
        model_type: Type[Message],
    ):
        self.aio_pika_message = aio_pika_message
        self.model_type = model_type

    def payload(self) -> Message:
        return self.model_type.model_validate_json(self.aio_pika_message.body)


class MessageProcessingLocker:

    def __init__(self) -> None:
        self.messages_lock = asyncio.Lock()
        self.current_processing_messages_set: set[asyncio.Task[Any]] = set()

    @asynccontextmanager
    async def lock_message_task(
        self, task: asyncio.Task[Any]
    ) -> AsyncGenerator[None, Any]:
        async with self.messages_lock:
            self.current_processing_messages_set.add(task)
            try:
                yield
            finally:
                self.current_processing_messages_set.discard(task)

    async def wait_all_messages_processed(self) -> None:
        if len(self.current_processing_messages_set) == 0:
            return

        await asyncio.gather(*self.current_processing_messages_set)


class MessageBusConsumer(ABC):

    async def consume(self) -> None:
        raise NotImplementedError("consume method not implemented")

    def shutdown(self) -> None: ...

    async def close(self) -> None:
        """Close all resources related to the consumer"""


class _WorkerShutdownState(ShutdownState):
    def __init__(self, shutdown_event: asyncio.Event):
        self.shutdown_event = shutdown_event

    def request_shutdown(self) -> None:
        self.shutdown_event.set()

    def is_shutdown_requested(self) -> bool:
        return self.shutdown_event.is_set()

    async def wait_for_shutdown(self) -> None:
        await self.shutdown_event.wait()


class AioPikaMicroserviceConsumer(MessageBusConsumer):
    def __init__(
        self,
        broker_backend: MessageBrokerBackend,
        config: AioPikaWorkerConfig,
        message_handler_set: MESSAGE_HANDLER_DATA_SET,
        scheduled_actions: SCHEDULED_ACTION_DATA_SET,
        uow_context_provider: UnitOfWorkContextProvider,
    ):

        self.broker_backend = broker_backend
        self.config = config
        self.message_handler_set = message_handler_set
        self.scheduled_actions = scheduled_actions
        self.incoming_map: dict[str, MessageHandlerData] = {}
        self.uow_context_provider = uow_context_provider
        self.shutdown_event = asyncio.Event()
        self.shutdown_state = _WorkerShutdownState(self.shutdown_event)
        self.lock = asyncio.Lock()
        self.tasks: set[asyncio.Task[Any]] = set()
        self.connection: aio_pika.abc.AbstractConnection | None = None
        self.channels: dict[str, aio_pika.abc.AbstractChannel] = {}

        # Connection resilience attributes
        self.connection_healthy = False
        self.connection_lock = asyncio.Lock()
        self.consumer_tags: dict[str, str] = {}  # Track consumer tags for cleanup
        self.health_check_task: asyncio.Task[Any] | None = None

    async def _verify_infrastructure(self) -> bool:
        """
        Verify that the required RabbitMQ infrastructure (exchanges, queues) exists.
        Returns True if all required infrastructure is in place.
        """
        try:
            async with self.connect() as connection:
                # Create a main channel just for checking infrastructure
                async with connection.channel() as main_channel:
                    # Get existing exchange and queues to verify infrastructure is in place
                    await RabbitmqUtils.get_main_exchange(
                        channel=main_channel,
                        exchange_name=self.config.exchange,
                    )
                    await RabbitmqUtils.get_dl_exchange(channel=main_channel)
                    await RabbitmqUtils.get_dl_queue(channel=main_channel)
                    return True
        except (ChannelNotFoundEntity, ChannelClosed, AMQPError) as e:
            logger.critical(
                f"Required exchange or queue infrastructure not found. "
                f"Please use the declare command first to create the required infrastructure. Error: {e}"
            )
            return False

    async def _setup_message_handler_consumer(
        self, handler: MessageHandlerData
    ) -> bool:
        """
        Set up a consumer for a message handler with retry mechanism.
        Returns True if successful, False otherwise.
        """
        queue_name = f"{handler.message_type.MESSAGE_TOPIC}.{handler.instance_callable.__module__}.{handler.instance_callable.__qualname__}"
        routing_key = f"{handler.message_type.MESSAGE_TOPIC}.#"

        async def setup_consumer() -> None:
            # Create a channel using the context manager
            async with self.create_channel(queue_name) as channel:
                queue: aio_pika.abc.AbstractQueue = await RabbitmqUtils.get_queue(
                    channel=channel, queue_name=queue_name
                )

                # Configure consumer and get the consumer tag
                consumer_tag = await queue.consume(
                    callback=MessageHandlerCallback(
                        consumer=self,
                        queue_name=queue_name,
                        routing_key=routing_key,
                        message_handler=handler,
                    ),
                    # no_ack=handler.spec.auto_ack,
                )

                # Store consumer tag for cleanup
                self.consumer_tags[queue_name] = consumer_tag

                logger.info(
                    "Consuming message handler %s on dedicated channel", queue_name
                )

                await self.shutdown_event.wait()

                logger.warning(
                    "Shutdown event received, stopping consumer for %s", queue_name
                )
                await queue.cancel(consumer_tag)

        try:
            # Setup with retry
            await retry_with_backoff(
                setup_consumer,
                retry_policy=self.config.consumer_retry_policy,
                retry_exceptions=(
                    ChannelNotFoundEntity,
                    ChannelClosed,
                    AMQPError,
                    AMQPConnectionError,
                    AMQPChannelError,
                    ConnectionClosed,
                ),
            )
            return True
        except Exception as e:
            logger.error(
                f"Failed to setup consumer for queue '{queue_name}' after retries: {e}"
            )
            return False

    async def _setup_scheduled_action_consumer(
        self, scheduled_action: ScheduledActionData
    ) -> bool:
        """
        Set up a consumer for a scheduled action with retry mechanism.
        Returns True if successful, False otherwise.
        """
        queue_name = f"{scheduled_action.callable.__module__}.{scheduled_action.callable.__qualname__}"
        routing_key = queue_name

        async def setup_consumer() -> None:
            # Create a channel using the context manager
            async with self.create_channel(queue_name) as channel:
                queue = await RabbitmqUtils.get_queue(
                    channel=channel, queue_name=queue_name
                )

                # Configure consumer and get the consumer tag
                consumer_tag = await queue.consume(
                    callback=ScheduledMessageHandlerCallback(
                        consumer=self,
                        queue_name=queue_name,
                        routing_key=routing_key,
                        scheduled_action=scheduled_action,
                    ),
                    no_ack=True,
                )

                # Store consumer tag for cleanup
                self.consumer_tags[queue_name] = consumer_tag

                logger.debug("Consuming scheduler %s on dedicated channel", queue_name)

                await self.shutdown_event.wait()

                logger.warning(
                    "Shutdown event received, stopping consumer for %s", queue_name
                )
                await queue.cancel(consumer_tag)

        try:
            # Setup with retry
            await retry_with_backoff(
                setup_consumer,
                retry_policy=self.config.consumer_retry_policy,
                retry_exceptions=(
                    ChannelNotFoundEntity,
                    ChannelClosed,
                    AMQPError,
                    AMQPConnectionError,
                    AMQPChannelError,
                    ConnectionClosed,
                ),
            )
            return True
        except Exception as e:
            logger.error(
                f"Failed to setup consumer for scheduler queue '{queue_name}' after retries: {e}"
            )
            return False

    async def consume(self) -> None:
        """
        Main consume method that sets up all message handlers and scheduled actions with retry mechanisms.
        """
        # Establish initial connection
        try:
            async with self.connect() as connection:
                self.connection_healthy = True

                # Start connection health monitoring
                self.health_check_task = asyncio.create_task(
                    self._monitor_connection_health(), name="ConnectionHealthMonitor"
                )

                # Verify infrastructure with retry
                infra_check_success = await retry_with_backoff(
                    self._verify_infrastructure,
                    retry_policy=self.config.connection_retry_config,
                    retry_exceptions=(Exception,),
                )

                if not infra_check_success:
                    logger.critical(
                        "Failed to verify RabbitMQ infrastructure. Shutting down."
                    )
                    self.shutdown_event.set()
                    return

                async def wait_for(
                    type: str, name: str, coroutine: Awaitable[bool]
                ) -> tuple[str, str, bool]:
                    return type, name, await coroutine

                tasks: set[asyncio.Task[tuple[str, str, bool]]] = set()

                # Setup message handlers
                for handler in self.message_handler_set:
                    queue_name = f"{handler.message_type.MESSAGE_TOPIC}.{handler.instance_callable.__module__}.{handler.instance_callable.__qualname__}"
                    self.incoming_map[queue_name] = handler

                    tasks.add(
                        task := asyncio.create_task(
                            wait_for(
                                "message_handler",
                                queue_name,
                                self._setup_message_handler_consumer(handler),
                            ),
                            name=f"MessageHandler-{queue_name}-setup-consumer",
                        )
                    )

                # Setup scheduled actions
                for scheduled_action in self.scheduled_actions:
                    queue_name = f"{scheduled_action.callable.__module__}.{scheduled_action.callable.__qualname__}"
                    tasks.add(
                        task := asyncio.create_task(
                            wait_for(
                                "scheduled_action",
                                queue_name,
                                self._setup_scheduled_action_consumer(scheduled_action),
                            ),
                            name=f"ScheduledAction-{queue_name}-setup-consumer",
                        )
                    )

                async def handle_task_results() -> None:
                    for task in asyncio.as_completed(tasks):
                        type, name, success = await task
                        if success:
                            logger.debug(
                                "Successfully set up %s consumer for %s", type, name
                            )
                        else:
                            logger.warning(
                                "Failed to set up %s consumer for %s, will not process messages from this queue",
                                type,
                                name,
                            )

                handle_task_results_task = asyncio.create_task(
                    handle_task_results(), name="HandleSetupTaskResults"
                )

                # Wait for shutdown signal
                await self.shutdown_event.wait()
                logger.debug("Shutdown event received, stopping consumers")

                await self.cancel_queue_consumers()

                # Cancel health monitoring
                if self.health_check_task:
                    self.health_check_task.cancel()
                    with suppress(asyncio.CancelledError):
                        await self.health_check_task

                handle_task_results_task.cancel()
                with suppress(asyncio.CancelledError):
                    await handle_task_results_task
                for task in tasks:
                    if not task.done():
                        task.cancel()
                        with suppress(asyncio.CancelledError):
                            await task
                logger.debug("Worker shutting down")
                # Wait for all tasks to complete
                await self.wait_all_tasks_done()

                # Close all channels and the connection
                await self.close_channels_and_connection()

        except Exception as e:
            logger.critical("Failed to establish initial connection to RabbitMQ: %s", e)
            # Re-raise the exception so it can be caught by the caller
            raise

    async def cancel_queue_consumers(self) -> None:
        """
        Cancel all active queue consumers.
        """
        logger.debug("Cancelling all active queue consumers...")
        for queue_name, channel in self.channels.items():
            try:
                if not channel.is_closed:
                    # Cancel consumer if we have its tag
                    if queue_name in self.consumer_tags:
                        try:
                            queue = await channel.get_queue(queue_name, ensure=False)
                            if queue:
                                await queue.cancel(self.consumer_tags[queue_name])
                        except Exception as cancel_error:
                            logger.warning(
                                "Error cancelling consumer for %s: %s",
                                queue_name,
                                cancel_error,
                            )
                        del self.consumer_tags[queue_name]
            except Exception as e:
                logger.warning("Error cancelling consumer for %s: %s", queue_name, e)

    async def wait_all_tasks_done(self) -> None:
        if not self.tasks:
            return

        logger.warning(
            "Waiting for (%s) in-flight tasks to complete: %s",
            len(self.tasks),
            ", ".join((task.get_name()) for task in self.tasks),
        )
        # async with self.lock:
        # Use gather with return_exceptions=True to ensure all tasks are awaited
        # even if some raise exceptions
        # results = await asyncio.gather(*self.tasks, return_exceptions=True)
        pending_tasks = [task for task in self.tasks if not task.done()]
        while len(pending_tasks) > 0:
            if not pending_tasks:
                break
            await asyncio.wait(pending_tasks, return_when=asyncio.FIRST_COMPLETED)

            pending_tasks = [task for task in pending_tasks if not task.done()]
            if len(pending_tasks) > 0:
                logger.warning(
                    "Waiting for (%s) in-flight tasks to complete: %s",
                    len(pending_tasks),
                    ", ".join((task.get_name()) for task in pending_tasks),
                )

        logger.warning("All in-flight tasks have completed.")
        # Log any exceptions that occurred
        # for result in results:
        #     if isinstance(result, Exception):
        #         logger.error("Task raised an exception during shutdown: %s", result)

    async def close_channels_and_connection(self) -> None:
        """Close all channels and then the connection"""
        logger.warning("Closing channels and connection...")
        await self._cleanup_connection()

    def shutdown(self) -> None:
        """Signal for shutdown"""
        logger.warning("Initiating graceful shutdown")
        self.shutdown_event.set()

    async def close(self) -> None:
        """Implement MessageBusConsumer.close for cleanup"""
        logger.warning("Closing consumer...")
        self.shutdown()

        # Cancel health monitoring
        if self.health_check_task:
            self.health_check_task.cancel()
            with suppress(asyncio.CancelledError):
                await self.health_check_task

        await self.wait_all_tasks_done()
        await self.close_channels_and_connection()

    async def get_channel(self, queue_name: str) -> aio_pika.abc.AbstractChannel | None:
        """
        Get the channel for a specific queue, or None if not found.
        This helps with error handling when a channel might have been closed.
        """
        if queue_name not in self.channels:
            logger.warning("No channel found for queue %s", queue_name)
            return None

        try:
            channel = self.channels[queue_name]
            if channel.is_closed:
                logger.warning("Channel for queue %s is closed", queue_name)
                # Remove the closed channel
                del self.channels[queue_name]

                # Attempt to recreate the channel if connection is healthy
                if (
                    self.connection
                    and not self.connection.is_closed
                    and self.connection_healthy
                ):
                    try:
                        logger.debug("Creating new channel for %s", queue_name)
                        self.channels[queue_name] = await self.connection.channel()
                        await self.channels[queue_name].set_qos(
                            prefetch_count=self.config.prefetch_count
                        )
                        return self.channels[queue_name]
                    except Exception as e:
                        logger.error(
                            "Failed to recreate channel for %s: %s", queue_name, e
                        )
                        # Trigger shutdown if channel creation fails
                        self._trigger_shutdown()
                        return None
                else:
                    # Connection is not healthy, trigger shutdown
                    self._trigger_shutdown()
                    return None
            return channel
        except Exception as e:
            logger.error("Error accessing channel for queue %s: %s", queue_name, e)
            # Trigger shutdown on any channel access error
            self._trigger_shutdown()
            return None

    async def _establish_channel(self, queue_name: str) -> aio_pika.abc.AbstractChannel:
        """
        Creates a new channel for the specified queue with proper QoS settings.
        """
        if self.connection is None or self.connection.is_closed:
            logger.warning(
                "Cannot create channel for %s: connection is not available", queue_name
            )
            raise RuntimeError("Connection is not available")

        logger.debug("Creating channel for queue %s", queue_name)
        channel = await self.connection.channel()
        await channel.set_qos(prefetch_count=self.config.prefetch_count)
        logger.debug("Created channel for queue %s", queue_name)
        return channel

    @asynccontextmanager
    async def create_channel(
        self, queue_name: str
    ) -> AsyncGenerator[aio_pika.abc.AbstractChannel, None]:
        """
        Create and yield a channel for the specified queue with retry mechanism.
        This context manager ensures the channel is properly managed.
        """
        try:
            # Create a new channel with retry
            channel = await retry_with_backoff(
                fn=lambda: self._establish_channel(queue_name),
                retry_policy=self.config.consumer_retry_policy,
                retry_exceptions=(
                    AMQPConnectionError,
                    AMQPChannelError,
                    ConnectionError,
                ),
            )

            # Save in the channels dict for tracking
            self.channels[queue_name] = channel
            logger.debug("Created new channel for queue %s", queue_name)

            try:
                yield channel
            finally:
                # Don't close the channel here as it might be used later
                # It will be closed during shutdown
                pass
        except aio_pika.exceptions.AMQPError as e:
            logger.error(
                "Error creating channel for queue %s after retries: %s", queue_name, e
            )
            raise

    async def _establish_connection(self) -> aio_pika.abc.AbstractConnection:
        """
        Creates a new RabbitMQ connection with retry logic.
        """
        try:
            logger.debug("Establishing connection to RabbitMQ")
            connection = await aio_pika.connect(
                self.config.url,
                heartbeat=self.config.connection_heartbeat_interval,
            )
            logger.debug("Connected to RabbitMQ successfully")
            return connection
        except Exception as e:
            logger.error("Failed to connect to RabbitMQ: %s", e)
            raise

    @asynccontextmanager
    async def connect(self) -> AsyncGenerator[aio_pika.abc.AbstractConnection, None]:
        """
        Create and manage the main connection to RabbitMQ with automatic retry.
        """
        if self.connection is not None and not self.connection.is_closed:
            logger.debug("Connection already exists, reusing existing connection")
            try:
                yield self.connection
            finally:
                # The existing connection will be handled by close_channels_and_connection
                pass
            return

        try:
            # Create a new connection with retry
            self.connection = await retry_with_backoff(
                self._establish_connection,
                retry_policy=self.config.connection_retry_config,
                retry_exceptions=(
                    AMQPConnectionError,
                    ConnectionError,
                    OSError,
                    TimeoutError,
                ),
            )

            try:
                yield self.connection
            finally:
                # Don't close the connection here; it will be closed in close_channels_and_connection
                pass
        except Exception as e:
            logger.error(
                "Failed to establish connection to RabbitMQ after retries: %s", e
            )
            if self.connection:
                try:
                    await self.connection.close()
                except Exception as close_error:
                    logger.error(
                        "Error closing connection after connect failure: %s",
                        close_error,
                    )
                self.connection = None
            raise

    @asynccontextmanager
    async def get_channel_ctx(
        self, queue_name: str
    ) -> AsyncGenerator[aio_pika.abc.AbstractChannel, None]:
        """
        Get a channel for a specific queue as a context manager.
        This is safer than using get_channel directly as it ensures proper error handling.
        """
        max_retries = 3
        retry_delay = 1.0

        for attempt in range(max_retries):
            try:
                channel = await self.get_channel(queue_name)
                if channel is not None:
                    try:
                        yield channel
                        return
                    finally:
                        # We don't close the channel here as it's managed by the consumer
                        pass

                # No channel available, check connection state
                if (
                    self.connection
                    and not self.connection.is_closed
                    and self.connection_healthy
                ):
                    # Try to create a new channel
                    async with self.create_channel(queue_name) as new_channel:
                        yield new_channel
                        return
                else:
                    # Connection is not healthy, trigger shutdown
                    logger.error(
                        "Connection not healthy while getting channel for %s, triggering shutdown",
                        queue_name,
                    )
                    self._trigger_shutdown()
                    raise RuntimeError(
                        f"Cannot get channel for queue {queue_name}: connection is not healthy"
                    )

            except Exception as e:
                if attempt < max_retries - 1:
                    logger.warning(
                        "Error getting channel for %s, retrying: %s", queue_name, e
                    )
                    await self._wait_delay_or_shutdown(retry_delay)
                    retry_delay *= 2
                else:
                    logger.error(
                        "Failed to get channel for %s after %s attempts: %s",
                        queue_name,
                        max_retries,
                        e,
                    )
                    raise

    async def _wait_delay_or_shutdown(self, delay: float) -> None:
        """
        Wait for the specified delay or exit early if shutdown is initiated.

        Args:
            delay: Delay in seconds to wait
        """

        wait_cor = asyncio.create_task(asyncio.sleep(delay), name="delayed-retry-wait")
        wait_shutdown_cor = asyncio.create_task(
            self.shutdown_event.wait(), name="delayed-retry-shutdown-wait"
        )

        await asyncio.wait(
            [wait_cor, wait_shutdown_cor],
            return_when=asyncio.FIRST_COMPLETED,
        )

    async def _monitor_connection_health(self) -> None:
        """
        Monitor connection health and trigger shutdown if connection is lost.
        This runs as a background task.
        """
        while not self.shutdown_event.is_set():
            try:
                await self._wait_delay_or_shutdown(
                    self.config.connection_health_check_interval
                )

                if self.shutdown_event.is_set():
                    break

                # Check connection health
                if not await self._is_connection_healthy():
                    logger.error(
                        "Connection health check failed, initiating worker shutdown"
                    )
                    self.shutdown()
                    break

            except asyncio.CancelledError:
                logger.debug("Connection health monitoring cancelled")
                break
            except Exception as e:
                logger.error("Error in connection health monitoring: %s", e)
                await self._wait_delay_or_shutdown(5)  # Wait before retrying

    async def _is_connection_healthy(self) -> bool:
        """
        Check if the connection is healthy.
        """
        try:
            if self.connection is None or self.connection.is_closed:
                return False

            # Try to create a temporary channel to test connection
            async with self.connection.channel() as test_channel:
                # If we can create a channel, connection is healthy
                return True

        except Exception as e:
            logger.debug("Connection health check failed: %s", e)
            return False

    def _trigger_shutdown(self) -> None:
        """
        Trigger worker shutdown due to connection loss.
        """
        if not self.shutdown_event.is_set():
            logger.error("Connection lost, initiating worker shutdown")
            self.connection_healthy = False
            self.shutdown()

    async def _cleanup_connection(self) -> None:
        """
        Clean up existing connection and channels.
        """
        # Cancel existing consumers
        for queue_name, channel in self.channels.items():
            try:
                if not channel.is_closed:
                    # Cancel consumer if we have its tag
                    if queue_name in self.consumer_tags:
                        try:
                            queue = await channel.get_queue(queue_name, ensure=False)
                            if queue:
                                await queue.cancel(self.consumer_tags[queue_name])
                        except Exception as cancel_error:
                            logger.warning(
                                "Error cancelling consumer for %s: %s",
                                queue_name,
                                cancel_error,
                            )
                        del self.consumer_tags[queue_name]
            except Exception as e:
                logger.warning("Error cancelling consumer for %s: %s", queue_name, e)

        # Close channels
        for queue_name, channel in self.channels.items():
            try:
                if not channel.is_closed:
                    await channel.close()
            except Exception as e:
                logger.warning("Error closing channel for %s: %s", queue_name, e)

        self.channels.clear()

        # Close connection
        if self.connection and not self.connection.is_closed:
            try:
                await self.connection.close()
            except Exception as e:
                logger.warning("Error closing connection: %s", e)

        self.connection = None
        self.connection_healthy = False


def create_message_bus(
    broker_url: str,
    broker_backend: MessageBrokerBackend,
    scheduled_actions: SCHEDULED_ACTION_DATA_SET,
    message_handler_set: MESSAGE_HANDLER_DATA_SET,
    uow_context_provider: UnitOfWorkContextProvider,
) -> MessageBusConsumer:

    parsed_url = urlparse(broker_url)

    if parsed_url.scheme == "amqp" or parsed_url.scheme == "amqps":
        assert parsed_url.query, "Query string must be set for AMQP URLs"

        query_params: dict[str, list[str]] = parse_qs(parsed_url.query)

        assert "exchange" in query_params, "Exchange must be set in the query string"
        assert (
            len(query_params["exchange"]) == 1
        ), "Exchange must be set in the query string"
        assert (
            "prefetch_count" in query_params
        ), "Prefetch count must be set in the query string"
        assert (
            len(query_params["prefetch_count"]) == 1
        ), "Prefetch count must be set in the query string"
        assert query_params["prefetch_count"][
            0
        ].isdigit(), "Prefetch count must be an integer in the query string"
        assert query_params["exchange"][0], "Exchange must be set in the query string"
        assert query_params["prefetch_count"][
            0
        ], "Prefetch count must be set in the query string"

        exchange = query_params["exchange"][0]
        prefetch_count = int(query_params["prefetch_count"][0])

        # Parse optional retry configuration parameters
        connection_retry_config = RetryPolicy()
        consumer_retry_config = RetryPolicy(
            max_retries=5, initial_delay=5, max_delay=60.0, backoff_factor=3.0
        )

        # Parse heartbeat and health check intervals
        connection_heartbeat_interval = 30.0
        connection_health_check_interval = 10.0

        # Connection retry config parameters
        if (
            "connection_retry_max" in query_params
            and query_params["connection_retry_max"][0].isdigit()
        ):
            connection_retry_config.max_retries = int(
                query_params["connection_retry_max"][0]
            )

        if "connection_retry_delay" in query_params:
            try:
                connection_retry_config.initial_delay = float(
                    query_params["connection_retry_delay"][0]
                )
            except ValueError:
                pass

        if "connection_retry_max_delay" in query_params:
            try:
                connection_retry_config.max_delay = float(
                    query_params["connection_retry_max_delay"][0]
                )
            except ValueError:
                pass

        if "connection_retry_backoff" in query_params:
            try:
                connection_retry_config.backoff_factor = float(
                    query_params["connection_retry_backoff"][0]
                )
            except ValueError:
                pass

        # Consumer retry config parameters
        if (
            "consumer_retry_max" in query_params
            and query_params["consumer_retry_max"][0].isdigit()
        ):
            consumer_retry_config.max_retries = int(
                query_params["consumer_retry_max"][0]
            )

        if "consumer_retry_delay" in query_params:
            try:
                consumer_retry_config.initial_delay = float(
                    query_params["consumer_retry_delay"][0]
                )
            except ValueError:
                pass

        if "consumer_retry_max_delay" in query_params:
            try:
                consumer_retry_config.max_delay = float(
                    query_params["consumer_retry_max_delay"][0]
                )
            except ValueError:
                pass

        if "consumer_retry_backoff" in query_params:
            try:
                consumer_retry_config.backoff_factor = float(
                    query_params["consumer_retry_backoff"][0]
                )
            except ValueError:
                pass

        # Heartbeat and health check intervals
        if "connection_heartbeat_interval" in query_params:
            try:
                connection_heartbeat_interval = float(
                    query_params["connection_heartbeat_interval"][0]
                )
            except ValueError:
                pass

        if "connection_health_check_interval" in query_params:
            try:
                connection_health_check_interval = float(
                    query_params["connection_health_check_interval"][0]
                )
            except ValueError:
                pass

        config = AioPikaWorkerConfig(
            url=broker_url,
            exchange=exchange,
            prefetch_count=prefetch_count,
            connection_retry_config=connection_retry_config,
            consumer_retry_policy=consumer_retry_config,
            connection_heartbeat_interval=connection_heartbeat_interval,
            connection_health_check_interval=connection_health_check_interval,
        )

        return AioPikaMicroserviceConsumer(
            config=config,
            broker_backend=broker_backend,
            message_handler_set=message_handler_set,
            scheduled_actions=scheduled_actions,
            uow_context_provider=uow_context_provider,
        )

    raise ValueError(
        f"Unsupported broker URL scheme: {parsed_url.scheme}. Supported schemes are amqp and amqps"
    )


class ScheduledMessageHandlerCallback:
    def __init__(
        self,
        consumer: AioPikaMicroserviceConsumer,
        queue_name: str,
        routing_key: str,
        scheduled_action: ScheduledActionData,
    ):
        self.consumer = consumer
        self.queue_name = queue_name
        self.routing_key = routing_key
        self.scheduled_action = scheduled_action

    async def __call__(
        self, aio_pika_message: aio_pika.abc.AbstractIncomingMessage
    ) -> None:

        if self.consumer.shutdown_event.is_set():
            logger.debug(
                "Shutdown in progress. Requeuing scheduled message for %s",
                self.queue_name,
            )
            try:
                # Use channel context for requeuing
                await aio_pika_message.reject(requeue=True)
            except RuntimeError:
                logger.warning(
                    "Could not requeue scheduled message during shutdown - channel not available"
                )
            except Exception as e:
                logger.error(
                    "Failed to requeue scheduled message during shutdown: %s", e
                )
            return

        # Check if connection is healthy before processing
        if not self.consumer.connection_healthy:
            logger.warning(
                "Connection not healthy, requeuing scheduled message for %s",
                self.queue_name,
            )
            try:
                if not self.consumer.connection_healthy:
                    # Still not healthy, requeue the message

                    await aio_pika_message.reject(requeue=True)
                    return
            except Exception as e:
                logger.error(
                    "Failed to requeue scheduled message due to connection issues: %s",
                    e,
                )
                return

        async with self.consumer.lock:
            task = asyncio.create_task(
                self.handle_message(aio_pika_message),
                name=f"ScheduledAction-{self.queue_name}-handle-message-{aio_pika_message.message_id}",
            )
            self.consumer.tasks.add(task)
            task.add_done_callback(self.handle_message_consume_done)

    def handle_message_consume_done(self, task: asyncio.Task[Any]) -> None:
        self.consumer.tasks.discard(task)
        if task.cancelled():
            logger.warning("Scheduled task for %s was cancelled", self.queue_name)
            return

        if (error := task.exception()) is not None:
            logger.exception(
                "Error processing scheduled action %s", self.queue_name, exc_info=error
            )

    async def handle_message(
        self, aio_pika_message: aio_pika.abc.AbstractIncomingMessage
    ) -> None:

        if self.consumer.shutdown_event.is_set():
            logger.debug(
                "Shutdown event set. Requeuing message for %s", self.queue_name
            )
            try:
                # Use channel context for requeuing

                await aio_pika_message.reject(requeue=True)
                return
            except RuntimeError:
                logger.warning(
                    "Could not requeue message during shutdown - channel not available"
                )
            except Exception as e:
                logger.error("Failed to requeue message during shutdown: %s", e)
                return

        # Check connection health before processing
        if not self.consumer.connection_healthy:
            logger.warning(
                "Connection not healthy, requeuing scheduled message for %s",
                self.queue_name,
            )
            try:

                await aio_pika_message.reject(requeue=True)
                return
            except Exception as e:
                logger.error(
                    "Failed to requeue scheduled message due to connection issues: %s",
                    e,
                )
                return

        sig = inspect.signature(self.scheduled_action.callable)
        if len(sig.parameters) == 1:

            task = asyncio.create_task(
                self.run_with_context(
                    self.scheduled_action,
                    (ScheduleDispatchData(int(aio_pika_message.body.decode("utf-8"))),),
                    {},
                ),
                name=f"ScheduledAction-{self.queue_name}-handle-message-{aio_pika_message.message_id}",
            )

        elif len(sig.parameters) == 0:
            task = asyncio.create_task(
                self.run_with_context(
                    self.scheduled_action,
                    (),
                    {},
                ),
                name=f"ScheduledAction-{self.queue_name}-handle-message-{aio_pika_message.message_id}",
            )
        else:
            logger.warning(
                "Scheduled action '%s' must have exactly one parameter of type ScheduleDispatchData or no parameters"
                % self.queue_name
            )
            return

        self.consumer.tasks.add(task)
        task.add_done_callback(self.handle_message_consume_done)

        try:
            await task
        except Exception as e:

            logger.exception(
                "Error processing scheduled action %s: %s", self.queue_name, e
            )

    async def run_with_context(
        self,
        scheduled_action: ScheduledActionData,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> None:

        with provide_shutdown_state(self.consumer.shutdown_state):
            async with self.consumer.uow_context_provider(
                AppTransactionContext(
                    controller_member_reflect=scheduled_action.controller_member,
                    transaction_data=SchedulerTransactionData(
                        task_name=scheduled_action.spec.name
                        or scheduled_action.callable.__qualname__,
                        scheduled_to=datetime.now(UTC),
                        cron_expression=scheduled_action.spec.cron,
                        triggered_at=datetime.now(UTC),
                    ),
                )
            ):

                await scheduled_action.callable(*args, **kwargs)


class MessageHandlerCallback:

    def __init__(
        self,
        consumer: AioPikaMicroserviceConsumer,
        queue_name: str,
        routing_key: str,
        message_handler: MessageHandlerData,
    ):
        self.consumer = consumer
        self.queue_name = queue_name
        self.routing_key = routing_key
        self.message_handler = message_handler
        self.retry_state: dict[str, dict[str, Any]] = {}

    async def message_consumer(
        self, aio_pika_message: aio_pika.abc.AbstractIncomingMessage
    ) -> None:
        if self.consumer.shutdown_event.is_set():
            logger.debug(
                "Shutdown in progress. Requeuing message for %s", self.queue_name
            )
            try:
                # Use channel context for requeuing

                await aio_pika_message.reject(requeue=True)
            except RuntimeError:
                logger.warning(
                    "Could not requeue message during shutdown - channel not available"
                )
            except Exception as e:
                logger.error("Failed to requeue message during shutdown: %s", e)
            return

        # Check if connection is healthy before processing
        if not self.consumer.connection_healthy:
            logger.warning(
                "Connection not healthy, requeuing message for %s", self.queue_name
            )
            try:
                if not self.consumer.connection_healthy:
                    # Still not healthy, requeue the message

                    await aio_pika_message.reject(requeue=True)
                    return
            except Exception as e:
                logger.error(
                    "Failed to requeue message due to connection issues: %s", e
                )
                return

        async with self.consumer.lock:
            task = asyncio.create_task(
                self.handle_message(aio_pika_message),
                name=f"MessageHandler-{self.queue_name}-handle-message-{aio_pika_message.message_id}",
            )
            self.consumer.tasks.add(task)

            def handle_message_consume_done(task: asyncio.Task[Any]) -> None:
                self.consumer.tasks.discard(task)
                if task.cancelled():
                    logger.warning("Task for queue %s was cancelled", self.queue_name)
                    return

                if (error := task.exception()) is not None:
                    logger.exception(
                        "Error processing message id %s for queue %s",
                        aio_pika_message.message_id,
                        self.queue_name,
                        exc_info=error,
                    )

            task.add_done_callback(handle_message_consume_done)

    async def __call__(
        self, aio_pika_message: aio_pika.abc.AbstractIncomingMessage
    ) -> None:
        await self.message_consumer(aio_pika_message)

    async def handle_reject_message(
        self,
        aio_pika_message: aio_pika.abc.AbstractIncomingMessage,
        *,
        requeue_timeout: float = 0,
        requeue: bool = False,
        retry_count: int = 0,
        exception: Optional[BaseException] = None,
    ) -> None:
        """
        Handle rejecting a message, with support for retry with exponential backoff.

        Args:
            aio_pika_message: The message to reject
            requeue: Whether to requeue the message directly (True) or handle with retry logic (False)
            retry_count: The current retry count for this message
            exception: The exception that caused the rejection, if any
        """
        message_id = aio_pika_message.message_id or str(uuid.uuid4())

        try:
            # Check if we should retry with backoff
            if (
                not requeue
                and self.message_handler.spec.nack_on_exception
                and exception is not None
            ):
                # Get retry config from consumer
                retry_config = (
                    self.message_handler.spec.retry_config
                    or self.consumer.config.consumer_retry_policy
                )

                # Check if we reached max retries
                if retry_count >= retry_config.max_retries:
                    logger.warning(
                        "Message %s (%s) failed after %s retries, dead-lettering: %s",
                        message_id,
                        self.queue_name,
                        retry_count,
                        str(exception),
                    )
                    # Dead-letter the message after max retries
                    try:
                        await aio_pika_message.reject(requeue=False)
                    except Exception as e:
                        logger.error(
                            "Failed to dead-letter message %s: %s", message_id, e
                        )
                    return

                # Calculate delay for this retry attempt
                delay = retry_config.initial_delay * (
                    retry_config.backoff_factor**retry_count
                )
                if retry_config.jitter:
                    jitter_amount = delay * 0.25
                    delay = delay + random.uniform(-jitter_amount, jitter_amount)
                    delay = max(
                        delay, 0.1
                    )  # Ensure delay doesn't go negative due to jitter

                delay = min(delay, retry_config.max_delay)

                logger.warning(
                    "Message %s (%s) failed with %s, retry %s/%s scheduled in %.2fs",
                    message_id,
                    self.queue_name,
                    str(exception),
                    retry_count + 1,
                    retry_config.max_retries,
                    delay,
                )

                # Store retry state for this message
                self.retry_state[message_id] = {
                    "retry_count": retry_count + 1,
                    "last_exception": exception,
                    "next_retry": time.time() + delay,
                }

                # Schedule retry after delay
                task = asyncio.create_task(
                    self._delayed_retry(
                        aio_pika_message, delay, retry_count + 1, exception
                    ),
                    name=f"MessageHandler-{self.queue_name}-delayed-retry-{message_id}",
                )
                self.consumer.tasks.add(task)

                # Acknowledge the current message since we'll handle retry ourselves
                try:
                    await aio_pika_message.ack()
                except Exception as e:
                    logger.error(
                        "Failed to acknowledge message %s for retry: %s", message_id, e
                    )
                return

            # Standard reject without retry or with immediate requeue
            try:
                await self._wait_delay_or_shutdown(
                    requeue_timeout
                )  # Optional delay before requeueing
                await aio_pika_message.reject(requeue=requeue)
                if requeue:
                    logger.warning(
                        "Message %s (%s) requeued for immediate retry",
                        message_id,
                        self.queue_name,
                    )
                else:
                    logger.warning(
                        "Message %s (%s) rejected without requeue",
                        message_id,
                        self.queue_name,
                    )
            except Exception as e:
                logger.error("Failed to reject message %s: %s", message_id, e)

        except Exception as e:
            logger.exception(
                "Unexpected error in handle_reject_message for %s (%s): %s",
                message_id,
                self.queue_name,
                e,
            )

    async def _delayed_retry(
        self,
        aio_pika_message: aio_pika.abc.AbstractIncomingMessage,
        delay: float,
        retry_count: int,
        exception: Optional[BaseException],
    ) -> None:
        """
        Handle delayed retry of a message after exponential backoff delay.

        Args:
            aio_pika_message: The original message
            delay: Delay in seconds before retrying
            retry_count: The current retry count (after increment)
            exception: The exception that caused the failure
        """
        message_id = aio_pika_message.message_id or str(uuid.uuid4())

        try:
            # Wait for the backoff delay
            await self._wait_delay_or_shutdown(delay)

            # Get message body and properties for republishing
            message_body = aio_pika_message.body
            headers = (
                aio_pika_message.headers.copy() if aio_pika_message.headers else {}
            )

            # Add retry information to headers
            headers["x-retry-count"] = retry_count
            if exception:
                headers["x-last-error"] = str(exception)

            # Clean up retry state
            if message_id in self.retry_state:
                del self.retry_state[message_id]

            # Republish the message to the same queue with retry logic
            max_attempts = 3
            for attempt in range(max_attempts):
                try:
                    async with self.consumer.get_channel_ctx(
                        self.queue_name
                    ) as channel:
                        exchange = await RabbitmqUtils.get_main_exchange(
                            channel=channel,
                            exchange_name=self.consumer.config.exchange,
                        )

                        await exchange.publish(
                            aio_pika.Message(
                                body=message_body,
                                headers=headers,
                                message_id=message_id,
                                content_type=aio_pika_message.content_type,
                                content_encoding=aio_pika_message.content_encoding,
                                delivery_mode=aio_pika_message.delivery_mode,
                            ),
                            routing_key=self.routing_key,
                        )

                        logger.warning(
                            "Message %s (%s) republished for retry %s",
                            message_id,
                            self.queue_name,
                            retry_count,
                        )
                        return

                except Exception as e:
                    if attempt < max_attempts - 1:
                        logger.warning(
                            "Failed to republish message %s (attempt %s): %s",
                            message_id,
                            attempt + 1,
                            e,
                        )
                        await self._wait_delay_or_shutdown(
                            (1.0 * (attempt + 1))
                        )  # Exponential backoff
                    else:
                        logger.error(
                            "Failed to republish message %s after %s attempts: %s",
                            message_id,
                            max_attempts,
                            e,
                        )
                        raise

        except Exception as e:
            logger.exception(
                "Failed to execute delayed retry for message %s (%s): %s",
                message_id,
                self.queue_name,
                e,
            )
            # If we fail to republish, try to dead-letter the original message
            try:
                if message_id in self.retry_state:
                    del self.retry_state[message_id]
            except Exception:
                pass

    async def _wait_delay_or_shutdown(self, delay: float) -> None:
        await self.consumer._wait_delay_or_shutdown(delay)

    async def handle_message(
        self, aio_pika_message: aio_pika.abc.AbstractIncomingMessage
    ) -> None:

        routing_key = self.queue_name

        if routing_key is None:
            logger.warning("No topic found for message")
            await self.handle_reject_message(aio_pika_message)
            return

        handler_data = self.message_handler

        handler = handler_data.instance_callable
        handler_method = handler_data.controller_member.member_function

        # sig = inspect.signature(handler)

        # if len(sig.parameters) != 1:
        #     logger.warning(
        #         "Handler for topic '%s' must have exactly one parameter which is MessageOf[T extends Message]"
        #         % routing_key
        #     )
        #     return

        # parameter = list(sig.parameters.values())[0]

        # param_origin = get_origin(parameter.annotation)

        # if param_origin is not MessageOf:
        #     logger.warning(
        #         "Handler for topic '%s' must have exactly one parameter of type Message"
        #         % routing_key
        #     )
        #     return

        # if len(parameter.annotation.__args__) != 1:
        #     logger.warning(
        #         "Handler for topic '%s' must have exactly one parameter of type Message"
        #         % routing_key
        #     )
        #     return

        # message_type = parameter.annotation.__args__[0]

        # if not issubclass(message_type, BaseModel):
        #     logger.warning(
        #         "Handler for topic '%s' must have exactly one parameter of type MessageOf[BaseModel]"
        #         % routing_key
        #     )
        #     return

        mode, message_type = MessageHandler.validate_decorated_fn(handler_method)

        built_message = AioPikaMessage(aio_pika_message, message_type)

        incoming_message_spec = MessageHandler.get_last(handler)
        assert incoming_message_spec is not None, "Incoming message spec must be set"
        # Extract retry count from headers if available
        headers = aio_pika_message.headers or {}
        retry_count = int(str(headers.get("x-retry-count", 0)))

        with provide_implicit_headers(aio_pika_message.headers), provide_shutdown_state(
            self.consumer.shutdown_state
        ):
            async with self.consumer.uow_context_provider(
                AppTransactionContext(
                    controller_member_reflect=handler_data.controller_member,
                    transaction_data=MessageBusTransactionData(
                        message_id=aio_pika_message.message_id,
                        processing_attempt=retry_count + 1,
                        message_type=message_type,
                        message=built_message,
                        topic=routing_key,
                    ),
                )
            ):
                maybe_timeout_ctx: AsyncContextManager[Any]
                if incoming_message_spec.timeout is not None:
                    maybe_timeout_ctx = asyncio.timeout(incoming_message_spec.timeout)
                else:
                    maybe_timeout_ctx = none_context()

                start_time = time.perf_counter()
                async with maybe_timeout_ctx:
                    try:
                        with provide_bus_message_controller(
                            AioPikaMessageBusController(aio_pika_message)
                        ):
                            try:
                                if mode == "WRAPPED":
                                    future = handler(built_message)
                                else:
                                    try:

                                        payload = built_message.payload()
                                    except ValidationError as exc:
                                        logger.exception(
                                            "Validation error parsing message %s on topic %s",
                                            aio_pika_message.message_id or "unknown",
                                            routing_key,
                                        )
                                        aio_pika_message.headers["x-last-error"] = (
                                            "Validation error parsing message payload"
                                        )
                                        await aio_pika_message.reject(requeue=False)
                                        record_exception(
                                            exc,
                                        )
                                        set_span_status("ERROR")
                                        return
                                    future = handler(payload)

                                await future

                                with suppress(aio_pika.MessageProcessError):
                                    # Use channel context for acknowledgement with retry
                                    try:
                                        await aio_pika_message.ack()
                                        set_span_status("OK")
                                    except Exception as ack_error:
                                        logger.warning(
                                            "Failed to acknowledge message %s: %s",
                                            aio_pika_message.message_id or "unknown",
                                            ack_error,
                                        )
                                successfully = True
                            except Exception as base_exc:
                                set_span_status("ERROR")
                                record_exception(
                                    base_exc,
                                    {
                                        "message_id": aio_pika_message.message_id
                                        or "unknown",
                                        "routing_key": routing_key,
                                    },
                                )
                                successfully = False
                                # Get message id for logging
                                message_id = aio_pika_message.message_id or "unknown"

                                # Process exception handler if configured
                                if incoming_message_spec.exception_handler is not None:
                                    try:
                                        incoming_message_spec.exception_handler(
                                            base_exc
                                        )
                                    except Exception as nested_exc:
                                        logger.exception(
                                            "Error processing exception handler for message %s: %s | %s",
                                            message_id,
                                            base_exc,
                                            nested_exc,
                                        )
                                else:
                                    logger.exception(
                                        "Error processing message %s on topic %s: %s",
                                        message_id,
                                        routing_key,
                                        str(base_exc),
                                    )

                                # Handle rejection with retry logic
                                if incoming_message_spec.nack_on_exception:
                                    await self.handle_reject_message(
                                        aio_pika_message,
                                        requeue_timeout=incoming_message_spec.nack_delay_on_exception,
                                        requeue=False,  # Don't requeue directly, use our backoff mechanism
                                        retry_count=retry_count,
                                        exception=base_exc,
                                    )
                                else:
                                    # Message shouldn't be retried, reject it
                                    await self.handle_reject_message(
                                        aio_pika_message,
                                        requeue=False,
                                        requeue_timeout=incoming_message_spec.nack_delay_on_exception,
                                        exception=base_exc,
                                    )

                            elapsed_time = time.perf_counter() - start_time
                            # Message processed successfully, log and clean up any retry state
                            message_id = aio_pika_message.message_id or str(
                                uuid.uuid4()
                            )
                            if message_id in self.retry_state:
                                del self.retry_state[message_id]

                            # Log success with retry information if applicable
                            headers = aio_pika_message.headers or {}
                            traceparent = headers.get("traceparent")
                            trace_info = (
                                f" [traceparent={str(traceparent)}]"
                                if traceparent
                                else ""
                            )

                            if "x-retry-count" in headers:
                                retry_count = int(str(headers.get("x-retry-count", 0)))
                                logger.debug(
                                    "Message %s#%s processed "
                                    + (
                                        "successfully"
                                        if successfully
                                        else "with errors"
                                    )
                                    + " after %s retries in %.4fs%s",
                                    message_id,
                                    self.queue_name,
                                    retry_count,
                                    elapsed_time,
                                    trace_info,
                                )
                            else:
                                logger.debug(
                                    "Message %s#%s processed "
                                    + (
                                        "successfully"
                                        if successfully
                                        else "with errors"
                                    )
                                    + " in %.4fs%s",
                                    message_id,
                                    self.queue_name,
                                    elapsed_time,
                                    trace_info,
                                )
                                ...

                    except Exception as base_exc:
                        logger.critical(
                            f"Critical error processing message {aio_pika_message.message_id} when providing bus message controller: {base_exc}"
                        )


@asynccontextmanager
async def none_context() -> AsyncGenerator[None, None]:
    yield


class MessageBusWorker:
    def __init__(
        self,
        app: Microservice,
        broker_url: str,
        backend_url: str,
        handler_names: set[str] | None = None,
    ) -> None:
        self.app = app
        self.backend_url = backend_url
        self.broker_url = broker_url
        self.handler_names = handler_names

        self.container = Container(app)
        self.lifecycle = AppLifecycle(app, self.container)

        self.uow_context_provider = UnitOfWorkContextProvider(
            app=app, container=self.container
        )

        self._consumer: MessageBusConsumer | None = None

    @property
    def consumer(self) -> MessageBusConsumer:
        if self._consumer is None:
            raise RuntimeError("Consumer not started")
        return self._consumer

    async def start_async(self) -> None:
        all_message_handlers_set: MESSAGE_HANDLER_DATA_SET = set()
        all_scheduled_actions_set: SCHEDULED_ACTION_DATA_SET = set()
        with providing_app_type("worker"):
            async with self.lifecycle():
                for instance_class in self.app.controllers:
                    controller = MessageBusController.get_last(instance_class)

                    if controller is None:
                        continue

                    instance: Any = self.container.get_by_type(instance_class)

                    factory = controller.get_messagebus_factory()
                    handlers, schedulers = factory(instance)

                    message_handler_data_map: dict[str, MessageHandlerData] = {}
                    all_scheduled_actions_set.update(schedulers)
                    for handler_data in handlers:
                        message_type = handler_data.spec.message_type
                        topic = message_type.MESSAGE_TOPIC

                        # Filter handlers by name if specified
                        if (
                            self.handler_names is not None
                            and handler_data.spec.name is not None
                        ):
                            if handler_data.spec.name not in self.handler_names:
                                continue
                        elif (
                            self.handler_names is not None
                            and handler_data.spec.name is None
                        ):
                            # Skip handlers without names when filtering is requested
                            continue

                        if (
                            topic in message_handler_data_map
                            and message_type.MESSAGE_TYPE == "task"
                        ):
                            logger.warning(
                                "Task handler for topic '%s' already registered. Skipping"
                                % topic
                            )
                            continue
                        message_handler_data_map[topic] = handler_data
                        all_message_handlers_set.add(handler_data)

                broker_backend = get_message_broker_backend_from_url(
                    url=self.backend_url
                )

                consumer = self._consumer = create_message_bus(
                    broker_url=self.broker_url,
                    broker_backend=broker_backend,
                    scheduled_actions=all_scheduled_actions_set,
                    message_handler_set=all_message_handlers_set,
                    uow_context_provider=self.uow_context_provider,
                )

                await consumer.consume()

    def start_sync(self) -> None:

        def on_shutdown(loop: asyncio.AbstractEventLoop) -> None:
            logger.warning("Shutting down - signal received")
            # Schedule the shutdown to run in the event loop
            asyncio.create_task(
                self._graceful_shutdown(), name="Worker-Graceful-Shutdown"
            )
            # wait until the shutdown is complete

        with asyncio.Runner(loop_factory=uvloop.new_event_loop) as runner:
            loop = runner.get_loop()
            loop.add_signal_handler(signal.SIGINT, on_shutdown, loop)
            # Add graceful shutdown handler for SIGTERM as well
            loop.add_signal_handler(signal.SIGTERM, on_shutdown, loop)
            try:
                runner.run(self.start_async())
            except Exception as e:
                logger.critical("Worker failed to start due to connection error: %s", e)
                # Exit with error code 1 to indicate startup failure
                import sys

                sys.exit(1)

    async def _graceful_shutdown(self) -> None:
        """Handles graceful shutdown process"""
        logger.warning("Initiating graceful shutdown sequence")
        # Use the comprehensive close method that handles shutdown, task waiting and connection cleanup

        self.consumer.shutdown()
        logger.warning("Graceful shutdown completed")


class AioPikaMessageBusController(BusMessageController):
    def __init__(self, aio_pika_message: aio_pika.abc.AbstractIncomingMessage):
        self.aio_pika_message = aio_pika_message
        # We access consumer callback through context if available
        self._callback: Optional[MessageHandlerCallback] = None

    def _get_callback(self) -> MessageHandlerCallback:
        """
        Find the callback associated with this message.
        This allows us to access the retry mechanisms.
        """
        if self._callback is None:
            # Get the context from current frame's locals
            frame = inspect.currentframe()
            if frame is not None:
                try:
                    caller_frame = frame.f_back
                    if caller_frame is not None:
                        # Check for context with handler callback
                        callback_ref = None
                        # Look for handler_message call context
                        while caller_frame is not None:
                            if "self" in caller_frame.f_locals:
                                self_obj = caller_frame.f_locals["self"]
                                if isinstance(self_obj, MessageHandlerCallback):
                                    callback_ref = self_obj
                                    break
                            caller_frame = caller_frame.f_back
                        # Save callback reference if we found it
                        self._callback = callback_ref
                finally:
                    del frame  # Avoid reference cycles

            if self._callback is None:
                raise RuntimeError("Could not find callback context for message retry")

        return self._callback

    async def ack(self) -> None:
        await self.aio_pika_message.ack()

    async def nack(self) -> None:
        await self.aio_pika_message.nack()

    async def reject(self) -> None:
        await self.aio_pika_message.reject()

    async def retry(self, delay: float = 5) -> None:
        """
        Retry the message immediately by rejecting with requeue flag.
        This doesn't use the exponential backoff mechanism.
        """
        callback = self._get_callback()
        await callback.handle_reject_message(
            self.aio_pika_message, requeue=True, requeue_timeout=delay
        )

    async def retry_later(self, delay: int) -> None:
        """
        Retry the message after a specified delay using the exponential backoff mechanism.

        Args:
            delay: Minimum delay in seconds before retrying
        """
        try:
            callback = self._get_callback()

            # Get current retry count from message headers
            headers = self.aio_pika_message.headers or {}
            retry_count = int(str(headers.get("x-retry-count", 0)))

            # Handle retry with explicit delay
            asyncio.create_task(
                callback._delayed_retry(
                    self.aio_pika_message,
                    float(delay),
                    retry_count + 1,
                    None,  # No specific exception
                ),
                name=f"MessageHandler-{callback.queue_name}-delayed-retry-{self.aio_pika_message.message_id or 'unknown'}-{int(time.time())}",
            )

            # Acknowledge the current message since we'll republish
            await self.aio_pika_message.ack()

        except Exception as e:
            logger.exception("Failed to schedule retry_later: %s", e)
            # Fall back to immediate retry
            await self.aio_pika_message.reject(requeue=True)



# SPDX-FileCopyrightText: 2025 Lucas S
#
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import ABC
from contextlib import asynccontextmanager
from typing import AsyncContextManager, AsyncGenerator, Iterable

from jararaca.scheduler.types import DelayedMessageData


class MessageBrokerBackend(ABC):

    def lock(self) -> AsyncContextManager[None]:
        """
        Acquire a lock for the message broker backend.
        This is used to ensure that only one instance of the scheduler is running at a time.
        """
        raise NotImplementedError(f"lock() is not implemented by {self.__class__}.")

    async def get_last_dispatch_time(self, action_name: str) -> int | None:
        """
        Get the last dispatch time of the scheduled action.
        This is used to determine if the scheduled action should be executed again
        or if it should be skipped.
        """
        raise NotImplementedError(
            f"get_last_dispatch_time() is not implemented by {self.__class__}."
        )

    async def set_last_dispatch_time(self, action_name: str, timestamp: int) -> None:
        """
        Set the last dispatch time of the scheduled action.
        This is used to determine if the scheduled action should be executed again
        or if it should be skipped.
        """
        raise NotImplementedError(
            f"set_last_dispatch_time() is not implemented by {self.__class__}."
        )

    async def get_in_execution_count(self, action_name: str) -> int:
        """
        Get the number of scheduled actions in execution.
        This is used to determine if the scheduled action should be executed again
        or if it should be skipped.
        """
        raise NotImplementedError(
            f"get_in_execution_count() is not implemented by {self.__class__}."
        )

    def in_execution(self, action_name: str) -> AsyncContextManager[None]:
        """
        Acquire a lock for the scheduled action.
        This is used to ensure that only one instance of the scheduled action is running at a time.
        """
        raise NotImplementedError(
            f"in_execution() is not implemented by {self.__class__}."
        )

    async def dequeue_next_delayed_messages(
        self, start_timestamp: int
    ) -> Iterable[DelayedMessageData]:
        """
        Dequeue the next delayed messages from the message broker.
        This is used to trigger the scheduled action.
        """
        raise NotImplementedError(
            f"dequeue_next_delayed_messages() is not implemented by {self.__class__}."
        )

    async def enqueue_delayed_message(
        self, delayed_message: DelayedMessageData
    ) -> None:
        """
        Enqueue a delayed message to the message broker.
        This is used to trigger the scheduled action.
        """
        raise NotImplementedError(
            f"enqueue_delayed_message() is not implemented by {self.__class__}."
        )

    async def dispose(self) -> None:
        """
        Dispose of the message broker backend.
        This is used to clean up resources used by the message broker backend.
        """


class NullBackend(MessageBrokerBackend):
    """
    A null backend that does nothing.
    This is used for testing purposes.
    """

    @asynccontextmanager
    async def lock(self) -> AsyncGenerator[None, None]:
        yield

    async def get_last_dispatch_time(self, action_name: str) -> int:
        return 0

    async def set_last_dispatch_time(self, action_name: str, timestamp: int) -> None:
        pass

    async def dispose(self) -> None:
        pass



# SPDX-FileCopyrightText: 2025 Lucas S
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import TypedDict, TypeVar

from jararaca.presentation.websocket.base_types import WebSocketMessageBase
from jararaca.reflect.decorators import StackableDecorator

DECORATED_CLASS = TypeVar("DECORATED_CLASS")


class WebSocketEndpointOptions(TypedDict): ...


class WebSocketEndpoint(StackableDecorator):

    ORDER_COUNTER = 0

    def __init__(self, path: str, options: WebSocketEndpointOptions = {}) -> None:
        self.path = path
        self.options = options
        WebSocketEndpoint.ORDER_COUNTER += 1
        self.order = WebSocketEndpoint.ORDER_COUNTER


INHERITS_WS_MESSAGE = TypeVar("INHERITS_WS_MESSAGE", bound=WebSocketMessageBase)


class RegisterWebSocketMessage(StackableDecorator):

    def __init__(self, *message_types: type[INHERITS_WS_MESSAGE]) -> None:
        self.message_types = message_types



# SPDX-FileCopyrightText: 2025 Lucas S
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
from contextlib import asynccontextmanager, contextmanager
from typing import TYPE_CHECKING, Any, AsyncGenerator, Generator, Literal, Protocol

from opentelemetry import metrics, trace
from opentelemetry._logs import set_logger_provider
from opentelemetry.baggage.propagation import W3CBaggagePropagator
from opentelemetry.exporter.otlp.proto.http._log_exporter import (
    OTLPLogExporter as LogExporter,
)
from opentelemetry.exporter.otlp.proto.http.metric_exporter import (
    OTLPMetricExporter as MeterExporter,
)
from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter as SpanExporter,
)
from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator

from jararaca.messagebus.implicit_headers import (
    ImplicitHeaders,
    provide_implicit_headers,
    use_implicit_headers,
)
from jararaca.microservice import (
    AppTransactionContext,
    Container,
    Microservice,
    use_app_transaction_context,
)
from jararaca.observability.constants import TRACEPARENT_KEY
from jararaca.observability.decorators import (
    AttributeMap,
    AttributeValue,
    TracingContextProvider,
    TracingContextProviderFactory,
    TracingSpan,
    TracingSpanContext,
)
from jararaca.observability.interceptor import ObservabilityProvider

if TYPE_CHECKING:
    from opentelemetry.trace import Span as _Span
    from typing_extensions import TypeIs

tracer: trace.Tracer = trace.get_tracer(__name__)


def extract_context_attributes(ctx: AppTransactionContext) -> dict[str, Any]:
    tx_data = ctx.transaction_data
    extra_attributes: dict[str, Any] = {}

    if tx_data.context_type == "http":
        extra_attributes = {
            "http.method": tx_data.request.method,
            "http.url": str(tx_data.request.url),
            "http.path": tx_data.request.url.path,
            "http.route.path": tx_data.request.scope["route"].path,
            "http.route.endpoint.name": tx_data.request["route"].endpoint.__qualname__,
            "http.query": tx_data.request.url.query,
            **{
                f"http.request.path_param.{k}": v
                for k, v in tx_data.request.path_params.items()
            },
            **{
                f"http.request.query_param.{k}": v
                for k, v in tx_data.request.query_params.items()
            },
            **{
                f"http.request.header.{k}": v
                for k, v in tx_data.request.headers.items()
            },
            "http.request.client.host": (
                tx_data.request.client.host if tx_data.request.client else ""
            ),
        }
    elif tx_data.context_type == "message_bus":
        extra_attributes = {
            "bus.message.id": tx_data.message_id,
            "bus.message.name": tx_data.message_type.__qualname__,
            "bus.message.module": tx_data.message_type.__module__,
            "bus.message.category": tx_data.message_type.MESSAGE_CATEGORY,
            "bus.message.type": tx_data.message_type.MESSAGE_TYPE,
            "bus.message.topic": tx_data.message_type.MESSAGE_TOPIC,
            "bus.message.processing_attempt": tx_data.processing_attempt,
        }
    elif tx_data.context_type == "websocket":
        extra_attributes = {
            "ws.url": str(tx_data.websocket.url),
        }
    elif tx_data.context_type == "scheduler":
        extra_attributes = {
            "sched.task_name": tx_data.task_name,
            "sched.scheduled_to": tx_data.scheduled_to.isoformat(),
            "sched.cron_expression": tx_data.cron_expression,
            "sched.triggered_at": tx_data.triggered_at.isoformat(),
        }
    return {
        "app.context_type": tx_data.context_type,
        "controller_member_reflect.rest_controller.class_name": ctx.controller_member_reflect.controller_reflect.controller_class.__qualname__,
        "controller_member_reflect.rest_controller.module": ctx.controller_member_reflect.controller_reflect.controller_class.__module__,
        "controller_member_reflect.member_function.name": ctx.controller_member_reflect.member_function.__qualname__,
        "controller_member_reflect.member_function.module": ctx.controller_member_reflect.member_function.__module__,
        **extra_attributes,
    }


class OtelTracingSpan(TracingSpan):

    def __init__(self, span: trace.Span) -> None:
        self.span = span


class OtelTracingSpanContext(TracingSpanContext):

    def __init__(self, span_context: trace.SpanContext) -> None:
        self.span_context = span_context


class OtelTracingContextProvider(TracingContextProvider):

    def __init__(self, app_context: AppTransactionContext) -> None:
        self.app_context = app_context

    @contextmanager
    def start_span_context(
        self,
        trace_name: str,
        context_attributes: AttributeMap | None,
    ) -> Generator[None, None, None]:

        with tracer.start_as_current_span(trace_name, attributes=context_attributes):
            yield

    def add_event(
        self, event_name: str, event_attributes: AttributeMap | None = None
    ) -> None:
        trace.get_current_span().add_event(name=event_name, attributes=event_attributes)

    def set_span_status(self, status_code: Literal["OK", "ERROR", "UNSET"]) -> None:
        span = trace.get_current_span()
        if status_code == "OK":
            span.set_status(trace.Status(trace.StatusCode.OK))
        elif status_code == "ERROR":
            span.set_status(trace.Status(trace.StatusCode.ERROR))
        else:
            span.set_status(trace.Status(trace.StatusCode.UNSET))

    def record_exception(
        self,
        exception: Exception,
        attributes: AttributeMap | None = None,
        escaped: bool = False,
    ) -> None:
        span = trace.get_current_span()
        span.record_exception(exception, attributes=attributes, escaped=escaped)

    def set_span_attribute(self, key: str, value: AttributeValue) -> None:
        span = trace.get_current_span()
        span.set_attribute(key, value)

    def update_span_name(self, new_name: str) -> None:
        span = trace.get_current_span()

        span.update_name(new_name)

    def add_link(self, span_context: TracingSpanContext) -> None:
        if not isinstance(span_context, OtelTracingSpanContext):
            return
        span = trace.get_current_span()
        span.add_link(span_context.span_context)

    def get_current_span(self) -> TracingSpan | None:
        return OtelTracingSpan(trace.get_current_span())

    def get_current_span_context(self) -> TracingSpanContext | None:
        return OtelTracingSpanContext(trace.get_current_span().get_span_context())


class OtelTracingContextProviderFactory(TracingContextProviderFactory):

    def provide_provider(
        self, app_context: AppTransactionContext
    ) -> TracingContextProvider:
        return OtelTracingContextProvider(app_context)

    @asynccontextmanager
    async def root_setup(
        self, app_context: AppTransactionContext
    ) -> AsyncGenerator[None, None]:

        title: str = "Unmapped App Context Execution"
        headers: dict[str, Any] = {}
        tx_data = app_context.transaction_data
        extra_attributes = extract_context_attributes(app_context)

        if tx_data.context_type == "http":
            headers = dict(tx_data.request.headers)
            title = f"HTTP {tx_data.request.method} {tx_data.request.url}"
            extra_attributes["http.request.body"] = (await tx_data.request.body())[
                :5000
            ].decode(errors="ignore")

        elif tx_data.context_type == "message_bus":
            title = f"Att#{tx_data.processing_attempt} Message Bus {tx_data.topic}"
            headers = use_implicit_headers() or {}

        elif tx_data.context_type == "websocket":
            headers = dict(tx_data.websocket.headers)
            title = f"WebSocket {tx_data.websocket.url}"

        elif tx_data.context_type == "scheduler":
            title = f"Scheduler Task {tx_data.task_name}"

        carrier = {
            key: value
            for key, value in headers.items()
            if key.lower().startswith("traceparent")
            or key.lower().startswith("tracestate")
        }

        ctx = TraceContextTextMapPropagator().extract(carrier)

        b2 = {
            key: value
            for key, value in headers.items()
            if key.lower().startswith("baggage")
        }

        ctx2 = W3CBaggagePropagator().extract(b2, context=ctx)

        with tracer.start_as_current_span(
            name=title,
            context=ctx2,
            attributes={
                **extra_attributes,
            },
        ) as root_span:
            cx = root_span.get_span_context()
            span_traceparent_id = hex(cx.trace_id)[2:].rjust(32, "0")
            if app_context.transaction_data.context_type == "http":
                app_context.transaction_data.request.scope[TRACEPARENT_KEY] = (
                    span_traceparent_id
                )
            elif app_context.transaction_data.context_type == "websocket":
                app_context.transaction_data.websocket.scope[TRACEPARENT_KEY] = (
                    span_traceparent_id
                )
            tracing_headers: ImplicitHeaders = {}
            TraceContextTextMapPropagator().inject(tracing_headers)
            W3CBaggagePropagator().inject(tracing_headers)
            with provide_implicit_headers(tracing_headers):
                yield


class LoggerHandlerCallback(Protocol):

    def __call__(self, logger_handler: logging.Handler) -> None: ...


class SpanWithName(Protocol):

    @property
    def name(self) -> str: ...


def is_span_with_name(span: Any) -> "TypeIs[SpanWithName]":
    return hasattr(span, "name")


class CustomLoggingHandler(LoggingHandler):

    def _translate(self, record: logging.LogRecord) -> dict[str, Any]:
        try:
            ctx = use_app_transaction_context()
            data = super()._translate(record)
            extra_attributes = extract_context_attributes(ctx)

            current_span: "_Span" = trace.get_current_span()

            data["attributes"] = {
                **data.get("attributes", {}),
                **extra_attributes,
                **(
                    {
                        "span_name": (
                            current_span.name if is_span_with_name(current_span) else ""
                        ),
                    }
                    if hasattr(current_span, "name")
                    and current_span.is_recording() is False
                    else {}
                ),
            }

            return data
        except LookupError:
            return super()._translate(record)


class OtelObservabilityProvider(ObservabilityProvider):

    def __init__(
        self,
        app_name: str,
        logs_exporter: LogExporter,
        span_exporter: SpanExporter,
        meter_exporter: MeterExporter,
        logging_handler_callback: LoggerHandlerCallback = lambda logger_handler: None,
        meter_export_interval: int = 5000,
    ) -> None:
        self.app_name = app_name
        self.logs_exporter = logs_exporter
        self.span_exporter = span_exporter
        self.meter_exporter = meter_exporter
        self.tracing_provider = OtelTracingContextProviderFactory()
        self.meter_export_interval = meter_export_interval
        self.logging_handler_callback = logging_handler_callback

    @asynccontextmanager
    async def setup(
        self, app: Microservice, container: Container
    ) -> AsyncGenerator[None, None]:
        ### Setup Resource

        resource = Resource(attributes={SERVICE_NAME: self.app_name})

        ### Setup Tracing
        provider = TracerProvider(resource=resource)

        trace.set_tracer_provider(provider)

        span_processor = BatchSpanProcessor(self.span_exporter)
        provider.add_span_processor(span_processor)

        ### Setup Logs
        logger_provider = LoggerProvider(resource=resource)

        set_logger_provider(logger_provider)

        logger_provider.add_log_record_processor(
            BatchLogRecordProcessor(self.logs_exporter)
        )

        logging_handler = CustomLoggingHandler(
            level=logging.DEBUG, logger_provider=logger_provider
        )

        # logging_handler.addFilter(lambda _: get_tracing_ctx_provider() is not None)

        self.logging_handler_callback(logging_handler)

        ### Setup Metrics
        metric_reader = PeriodicExportingMetricReader(
            self.meter_exporter, export_interval_millis=self.meter_export_interval
        )
        meter_provider = MeterProvider(metric_readers=[metric_reader])

        metrics.set_meter_provider(meter_provider)

        yield

    @staticmethod
    def from_url(
        app_name: str,
        url: str,
        logging_handler_callback: LoggerHandlerCallback = lambda logger_handler: None,
        meter_export_interval: int = 5000,
    ) -> "OtelObservabilityProvider":
        """
        Create an instance of OtelObservabilityProvider with Http Exporters from a given URL
        """

        logs_exporter = LogExporter(endpoint=f"{url}/v1/logs")
        span_exporter = SpanExporter(endpoint=f"{url}/v1/traces")
        metric_exporter = MeterExporter(endpoint=f"{url}/v1/metrics")

        return OtelObservabilityProvider(
            app_name,
            logs_exporter,
            span_exporter,
            metric_exporter,
            logging_handler_callback,
            meter_export_interval,
        )



# SPDX-FileCopyrightText: 2025 Lucas S
#
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime, tzinfo
from typing import Generic, Protocol, TypeVar

from jararaca.messagebus.publisher import IMessage, use_publisher


class Message(IMessage):

    async def publish(self) -> None:
        task_publisher = use_publisher()
        await task_publisher.publish(self, self.MESSAGE_TOPIC)

    async def delay(self, seconds: int) -> None:
        task_publisher = use_publisher()
        await task_publisher.delay(self, seconds)

    async def schedule(self, when: datetime, tz: tzinfo) -> None:
        task_publisher = use_publisher()
        await task_publisher.schedule(self, when, tz)


INHERITS_MESSAGE_CO = TypeVar("INHERITS_MESSAGE_CO", bound=Message, covariant=True)


class MessageOf(Protocol, Generic[INHERITS_MESSAGE_CO]):

    def payload(self) -> INHERITS_MESSAGE_CO: ...



# SPDX-FileCopyrightText: 2025 Lucas S
#
# SPDX-License-Identifier: GPL-3.0-or-later

import asyncio
import contextlib
import logging
import signal
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import UTC, datetime
from typing import Any
from urllib.parse import parse_qs

import aio_pika
import croniter
import urllib3
import urllib3.util
import uvloop
from aio_pika import connect
from aio_pika.abc import AbstractChannel, AbstractConnection
from aio_pika.exceptions import (
    AMQPChannelError,
    AMQPConnectionError,
    AMQPError,
    ChannelClosed,
    ConnectionClosed,
)
from aio_pika.pool import Pool
from aiormq.exceptions import ChannelInvalidStateError

from jararaca.broker_backend import MessageBrokerBackend
from jararaca.broker_backend.mapper import get_message_broker_backend_from_url
from jararaca.core.uow import UnitOfWorkContextProvider
from jararaca.di import Container
from jararaca.lifecycle import AppLifecycle
from jararaca.microservice import Microservice, providing_app_type
from jararaca.scheduler.decorators import (
    ScheduledAction,
    ScheduledActionData,
    get_type_scheduled_actions,
)
from jararaca.scheduler.types import DelayedMessageData
from jararaca.utils.rabbitmq_utils import RabbitmqUtils
from jararaca.utils.retry import RetryPolicy, retry_with_backoff

logger = logging.getLogger(__name__)


def _extract_scheduled_actions(
    app: Microservice, container: Container, scheduler_names: set[str] | None = None
) -> list[ScheduledActionData]:
    scheduled_actions: list[ScheduledActionData] = []
    for controllers in app.controllers:

        controller_instance: Any = container.get_by_type(controllers)

        controller_scheduled_actions = get_type_scheduled_actions(controller_instance)

        # Filter scheduled actions by name if scheduler_names is provided
        if scheduler_names is not None:
            filtered_actions = []
            for action in controller_scheduled_actions:
                # Include actions that have a name and it's in the provided set
                if action.spec.name and action.spec.name in scheduler_names:
                    filtered_actions.append(action)
                # Skip actions without names when filtering is active
            controller_scheduled_actions = filtered_actions

        scheduled_actions.extend(controller_scheduled_actions)

    return scheduled_actions


# region Message Broker Dispatcher


class _MessageBrokerDispatcher(ABC):

    @abstractmethod
    async def dispatch_scheduled_action(
        self,
        action_id: str,
        timestamp: int,
    ) -> None:
        """
        Dispatch a message to the message broker.
        This is used to send a message to the message broker
        to trigger the scheduled action.
        """
        raise NotImplementedError("dispatch() is not implemented yet.")

    @abstractmethod
    async def dispatch_delayed_message(
        self,
        delayed_message: DelayedMessageData,
    ) -> None:
        """
        Dispatch a delayed message to the message broker.
        This is used to send a message to the message broker
        to trigger the scheduled action.
        """

        raise NotImplementedError("dispatch_delayed_message() is not implemented yet.")

    @abstractmethod
    async def initialize(self, scheduled_actions: list[ScheduledActionData]) -> None:
        raise NotImplementedError("initialize() is not implemented yet.")

    async def dispose(self) -> None:
        pass


class _RabbitMQBrokerDispatcher(_MessageBrokerDispatcher):

    def __init__(
        self,
        url: str,
        config: "BeatWorkerConfig | None" = None,
        shutdown_event: asyncio.Event | None = None,
    ) -> None:
        self.url = url
        self.config = config or BeatWorkerConfig()
        self.connection_healthy = False
        self.shutdown_event = shutdown_event or asyncio.Event()
        self.health_check_task: asyncio.Task[Any] | None = None

        self.conn_pool: "Pool[AbstractConnection]" = Pool(
            self._create_connection,
            max_size=self.config.max_pool_size,
        )

        self.channel_pool: "Pool[AbstractChannel]" = Pool(
            self._create_channel,
            max_size=self.config.max_pool_size,
        )

        splitted = urllib3.util.parse_url(url)

        assert splitted.scheme in ["amqp", "amqps"], "Invalid URL scheme"

        assert splitted.host, "Invalid URL host"

        assert splitted.query, "Invalid URL query"

        query_params: dict[str, list[str]] = parse_qs(splitted.query)

        assert "exchange" in query_params, "Missing exchange parameter"

        assert query_params["exchange"], "Empty exchange parameter"

        self.exchange = str(query_params["exchange"][0])

    async def _create_connection(self) -> AbstractConnection:
        """
        Create a connection to the RabbitMQ server with retry logic.
        """

        async def _establish_connection() -> AbstractConnection:
            logger.debug("Establishing connection to RabbitMQ")
            connection = await connect(
                self.url,
                heartbeat=self.config.connection_heartbeat_interval,
            )
            logger.debug("Connected to RabbitMQ successfully")
            return connection

        return await retry_with_backoff(
            _establish_connection,
            retry_policy=self.config.connection_retry_config,
            retry_exceptions=(
                AMQPConnectionError,
                ConnectionError,
                OSError,
                TimeoutError,
            ),
        )

    async def _create_channel(self) -> AbstractChannel:
        """
        Create a channel to the RabbitMQ server with retry logic.
        """

        async def _establish_channel() -> AbstractChannel:
            async with self.conn_pool.acquire() as connection:
                channel = await connection.channel()
                return channel

        return await retry_with_backoff(
            _establish_channel,
            retry_policy=self.config.connection_retry_config,
            retry_exceptions=(
                AMQPConnectionError,
                AMQPChannelError,
                ChannelClosed,
                ConnectionError,
            ),
        )

    async def dispatch_scheduled_action(self, action_id: str, timestamp: int) -> None:
        """
        Dispatch a message to the RabbitMQ server with retry logic.
        """
        if not self.connection_healthy:
            await self._wait_for_connection()

        async def _dispatch() -> None:
            logger.debug("Dispatching message to %s at %s", action_id, timestamp)
            async with self.channel_pool.acquire() as channel:
                exchange = await RabbitmqUtils.get_main_exchange(channel, self.exchange)

                await exchange.publish(
                    aio_pika.Message(body=str(timestamp).encode()),
                    routing_key=action_id,
                )
                logger.debug("Dispatched message to %s at %s", action_id, timestamp)

        try:
            await retry_with_backoff(
                _dispatch,
                retry_policy=self.config.dispatch_retry_config,
                retry_exceptions=(
                    AMQPConnectionError,
                    AMQPChannelError,
                    ChannelClosed,
                    ConnectionClosed,
                    AMQPError,
                ),
            )

        except ChannelInvalidStateError as e:
            logger.error(
                "Channel invalid state error when dispatching to %s: %s", action_id, e
            )
            # Trigger shutdown if dispatch fails
            self.shutdown_event.set()
            raise

        except Exception as e:
            logger.error(
                "Failed to dispatch message to %s after retries: %s", action_id, e
            )
            # Trigger shutdown if dispatch fails
            self.shutdown_event.set()
            raise

    async def dispatch_delayed_message(
        self, delayed_message: DelayedMessageData
    ) -> None:
        """
        Dispatch a delayed message to the RabbitMQ server with retry logic.
        """
        if not self.connection_healthy:
            await self._wait_for_connection()

        async def _dispatch() -> None:
            async with self.channel_pool.acquire() as channel:
                exchange = await RabbitmqUtils.get_main_exchange(channel, self.exchange)
                await exchange.publish(
                    aio_pika.Message(
                        body=delayed_message.payload,
                    ),
                    routing_key=f"{delayed_message.message_topic}.#",
                )

        try:
            await retry_with_backoff(
                _dispatch,
                retry_policy=self.config.dispatch_retry_config,
                retry_exceptions=(
                    AMQPConnectionError,
                    AMQPChannelError,
                    ChannelClosed,
                    ConnectionClosed,
                    AMQPError,
                ),
            )
        except Exception as e:
            logger.error("Failed to dispatch delayed message after retries: %s", e)
            # Trigger shutdown if dispatch fails
            self.shutdown_event.set()
            raise

    async def initialize(self, scheduled_actions: list[ScheduledActionData]) -> None:
        """
        Initialize the RabbitMQ server with retry logic.
        """

        async def _initialize() -> None:
            async with self.channel_pool.acquire() as channel:
                await RabbitmqUtils.get_main_exchange(channel, self.exchange)

                for sched_act_data in scheduled_actions:
                    queue_name = ScheduledAction.get_function_id(
                        sched_act_data.callable
                    )

                    # Try to get existing queue
                    await RabbitmqUtils.get_scheduled_action_queue(
                        channel=channel,
                        queue_name=queue_name,
                    )

        try:
            logger.debug("Initializing RabbitMQ connection...")
            await retry_with_backoff(
                _initialize,
                retry_policy=self.config.connection_retry_config,
                retry_exceptions=(
                    AMQPConnectionError,
                    AMQPChannelError,
                    ChannelClosed,
                    ConnectionClosed,
                    AMQPError,
                ),
            )

            # Verify connection is actually healthy after initialization
            if not await self._is_connection_healthy():
                logger.warning(
                    "Connection health check failed after initialization, retrying..."
                )
                # Wait a bit and try again
                await asyncio.sleep(2.0)
                if not await self._is_connection_healthy():
                    raise ConnectionError("Connection not healthy after initialization")

            self.connection_healthy = True
            logger.debug("RabbitMQ connection initialized successfully")

            # Start health monitoring
            self.health_check_task = asyncio.create_task(
                self._monitor_connection_health()
            )

        except Exception as e:
            logger.error("Failed to initialize RabbitMQ after retries: %s", e)
            raise

    async def dispose(self) -> None:
        """Clean up resources"""
        logger.debug("Disposing RabbitMQ broker dispatcher")
        self.shutdown_event.set()

        # Cancel health monitoring
        if self.health_check_task:
            self.health_check_task.cancel()
            try:
                await self.health_check_task
            except asyncio.CancelledError:
                pass

        # Clean up pools
        await self._cleanup_pools()

    async def _monitor_connection_health(self) -> None:
        """Monitor connection health and trigger shutdown if needed"""
        while not self.shutdown_event.is_set():
            try:
                await asyncio.sleep(self.config.health_check_interval)

                if self.shutdown_event.is_set():
                    break

                # Check connection health
                if not await self._is_connection_healthy():
                    logger.error("Connection health check failed, triggering shutdown")
                    self.shutdown_event.set()
                    break

            except asyncio.CancelledError:
                logger.debug("Connection health monitoring cancelled")
                break
            except Exception as e:
                logger.error("Error in connection health monitoring: %s", e)
                await asyncio.sleep(5)  # Wait before retrying

    async def _is_connection_healthy(self) -> bool:
        """Check if the connection is healthy"""
        try:
            # Try to acquire a connection from the pool
            async with self.conn_pool.acquire() as connection:
                if connection.is_closed:
                    return False

                # Try to create a channel to test connection
                channel = await connection.channel()
                await channel.close()
                return True

        except Exception as e:
            logger.debug("Connection health check failed: %s", e)
            return False

    async def _cleanup_pools(self) -> None:
        """Clean up existing connection pools"""
        try:
            if hasattr(self, "channel_pool"):
                await self.channel_pool.close()
        except Exception as e:
            logger.warning("Error closing channel pool: %s", e)

        try:
            if hasattr(self, "conn_pool"):
                await self.conn_pool.close()
        except Exception as e:
            logger.warning("Error closing connection pool: %s", e)

    async def _wait_for_connection(self) -> None:
        """Wait for connection to be healthy"""
        max_wait = 30.0  # Maximum wait time
        wait_time = 0.0

        while not self.connection_healthy and wait_time < max_wait:
            if self.shutdown_event.is_set():
                raise ConnectionError("Shutdown requested while waiting for connection")

            await asyncio.sleep(0.5)
            wait_time += 0.5

        if not self.connection_healthy:
            raise ConnectionError("Connection not healthy after maximum wait time")


def _get_message_broker_dispatcher_from_url(
    url: str,
    config: "BeatWorkerConfig | None" = None,
    shutdown_event: asyncio.Event | None = None,
) -> _MessageBrokerDispatcher:
    """
    Factory function to create a message broker instance from a URL.
    Currently, only RabbitMQ is supported.
    """
    if url.startswith("amqp://") or url.startswith("amqps://"):
        return _RabbitMQBrokerDispatcher(
            url=url, config=config, shutdown_event=shutdown_event
        )
    else:
        raise ValueError(f"Unsupported message broker URL: {url}")


# endregion


@dataclass
class BeatWorkerConfig:
    """Configuration for beat worker connection resilience"""

    connection_retry_config: RetryPolicy = field(
        default_factory=lambda: RetryPolicy(
            max_retries=10,
            initial_delay=2.0,
            max_delay=60.0,
            backoff_factor=2.0,
            jitter=True,
        )
    )
    dispatch_retry_config: RetryPolicy = field(
        default_factory=lambda: RetryPolicy(
            max_retries=3,
            initial_delay=1.0,
            max_delay=10.0,
            backoff_factor=2.0,
            jitter=True,
        )
    )
    connection_heartbeat_interval: float = 30.0
    health_check_interval: float = 15.0

    # Connection establishment timeouts
    connection_wait_timeout: float = 300.0  # 5 minutes to wait for initial connection

    # Pool configuration
    max_pool_size: int = 10
    pool_recycle_time: float = 3600.0  # 1 hour


class BeatWorker:

    def __init__(
        self,
        app: Microservice,
        interval: int,
        broker_url: str,
        backend_url: str,
        scheduled_action_names: set[str] | None = None,
        config: "BeatWorkerConfig | None" = None,
        shutdown_event: asyncio.Event | None = None,
    ) -> None:
        self.shutdown_event = shutdown_event or asyncio.Event()

        self.app = app
        self.config = config or BeatWorkerConfig()

        self.broker: _MessageBrokerDispatcher = _get_message_broker_dispatcher_from_url(
            broker_url, self.config, shutdown_event=self.shutdown_event
        )
        self.backend: MessageBrokerBackend = get_message_broker_backend_from_url(
            backend_url
        )

        self.interval = interval
        self.scheduler_names = scheduled_action_names
        self.container = Container(self.app)
        self.uow_provider = UnitOfWorkContextProvider(app, self.container)

        self.lifecycle = AppLifecycle(app, self.container)

    def run(self) -> None:

        def on_shutdown(loop: asyncio.AbstractEventLoop) -> None:
            logger.debug("Shutting down - signal received")
            # Schedule the shutdown to run in the event loop
            asyncio.create_task(self._graceful_shutdown())

        with asyncio.Runner(loop_factory=uvloop.new_event_loop) as runner:
            loop = runner.get_loop()
            loop.add_signal_handler(signal.SIGINT, on_shutdown, loop)
            # Add graceful shutdown handler for SIGTERM as well
            loop.add_signal_handler(signal.SIGTERM, on_shutdown, loop)
            try:
                runner.run(self.start_scheduler())
            except Exception as e:
                logger.critical(
                    "Scheduler failed to start due to connection error: %s", e
                )
                # Exit with error code 1 to indicate startup failure
                import sys

                sys.exit(1)

    async def start_scheduler(self) -> None:
        """
        Declares the scheduled actions and starts the scheduler.
        This is the main entry point for the scheduler.
        """
        with providing_app_type("beat"):
            async with self.lifecycle():

                scheduled_actions = _extract_scheduled_actions(
                    self.app, self.container, self.scheduler_names
                )

                # Initialize and wait for connection to be established
                logger.debug("Initializing broker connection...")
                await self.broker.initialize(scheduled_actions)

                # Wait for connection to be healthy before starting scheduler
                logger.debug("Waiting for connection to be established...")
                await self._wait_for_broker_connection()

                logger.debug("Connection established, starting scheduler...")
                await self.run_scheduled_actions(scheduled_actions)

    async def run_scheduled_actions(
        self, scheduled_actions: list[ScheduledActionData]
    ) -> None:

        logger.debug("Starting scheduled actions processing loop")

        # Ensure we have a healthy connection before starting the main loop
        if (
            isinstance(self.broker, _RabbitMQBrokerDispatcher)
            and not self.broker.connection_healthy
        ):
            logger.error("Connection not healthy at start of processing loop. Exiting.")
            return

        while not self.shutdown_event.is_set():
            # Check connection health before processing scheduled actions
            if (
                isinstance(self.broker, _RabbitMQBrokerDispatcher)
                and not self.broker.connection_healthy
            ):
                logger.error("Broker connection is not healthy. Exiting.")
                break

            now = int(time.time())
            for sched_act_data in scheduled_actions:
                func = sched_act_data.callable
                scheduled_action = sched_act_data.spec
                if self.shutdown_event.is_set():
                    break

                try:
                    async with self.backend.lock():

                        last_dispatch_time: int | None = (
                            await self.backend.get_last_dispatch_time(
                                ScheduledAction.get_function_id(func)
                            )
                        )

                        if last_dispatch_time is not None:
                            cron = croniter.croniter(
                                scheduled_action.cron, last_dispatch_time
                            )
                            next_run: datetime = cron.get_next(datetime).replace(
                                tzinfo=UTC
                            )
                            if next_run > datetime.now(UTC):
                                logger.debug(
                                    "Skipping %s.%s until %s",
                                    func.__module__,
                                    func.__qualname__,
                                    next_run,
                                )
                                continue

                        if not scheduled_action.allow_overlap:
                            if (
                                await self.backend.get_in_execution_count(
                                    ScheduledAction.get_function_id(func)
                                )
                                > 0
                            ):
                                continue

                        try:
                            start_time = time.perf_counter()
                            await self.broker.dispatch_scheduled_action(
                                ScheduledAction.get_function_id(func),
                                now,
                            )
                            elapsed_time = time.perf_counter() - start_time

                            await self.backend.set_last_dispatch_time(
                                ScheduledAction.get_function_id(func), now
                            )

                            logger.debug(
                                "Scheduled %s.%s at %s in %.4fs",
                                func.__module__,
                                func.__qualname__,
                                now,
                                elapsed_time,
                            )
                        except ChannelInvalidStateError as e:
                            logger.error(
                                "Channel invalid state error when dispatching %s.%s: %s",
                                func.__module__,
                                func.__qualname__,
                                e,
                            )
                            # Trigger shutdown if dispatch fails
                            self.shutdown_event.set()
                            raise
                        except Exception as e:
                            logger.error(
                                "Failed to dispatch scheduled action %s.%s: %s",
                                func.__module__,
                                func.__qualname__,
                                e,
                            )
                            # Continue with other scheduled actions even if one fails
                            continue

                except Exception as e:
                    logger.error(
                        "Error processing scheduled action %s.%s: %s",
                        func.__module__,
                        func.__qualname__,
                        e,
                    )
                    # Continue with other scheduled actions even if one fails
                    continue

            # Handle delayed messages
            try:
                delayed_messages = await self.backend.dequeue_next_delayed_messages(now)
                for delayed_message_data in delayed_messages:
                    try:
                        start_time = time.perf_counter()
                        await self.broker.dispatch_delayed_message(delayed_message_data)
                        elapsed_time = time.perf_counter() - start_time
                        logger.debug(
                            "Dispatched delayed message for topic %s in %.4fs",
                            delayed_message_data.message_topic,
                            elapsed_time,
                        )
                    except Exception as e:
                        logger.error("Failed to dispatch delayed message: %s", e)
                        # Continue with other delayed messages even if one fails
                        continue
            except Exception as e:
                logger.error("Error processing delayed messages: %s", e)

            with contextlib.suppress(asyncio.TimeoutError):
                await asyncio.wait_for(self.shutdown_event.wait(), self.interval)

        logger.debug("Scheduler stopped")

        try:
            await self.backend.dispose()
        except Exception as e:
            logger.error("Error disposing backend: %s", e)

        try:
            await self.broker.dispose()
        except Exception as e:
            logger.error("Error disposing broker: %s", e)

    async def _graceful_shutdown(self) -> None:
        """Handles graceful shutdown process"""
        logger.debug("Initiating graceful shutdown sequence")
        self.shutdown_event.set()
        logger.debug("Graceful shutdown completed")

    async def _wait_for_broker_connection(self) -> None:
        """
        Wait for the broker connection to be established and healthy.
        This ensures the scheduler doesn't start until RabbitMQ is ready.
        """
        max_wait_time = self.config.connection_wait_timeout
        check_interval = 2.0  # Check every 2 seconds
        elapsed_time = 0.0

        logger.debug(
            "Waiting for broker connection to be established (timeout: %ss)...",
            max_wait_time,
        )

        while elapsed_time < max_wait_time:
            if self.shutdown_event.is_set():
                raise ConnectionError(
                    "Shutdown requested while waiting for broker connection"
                )

            # Check if broker connection is healthy
            if (
                isinstance(self.broker, _RabbitMQBrokerDispatcher)
                and self.broker.connection_healthy
            ):
                logger.debug("Broker connection is healthy")
                return

            # If broker doesn't have health status, try a simple health check
            if not hasattr(self.broker, "connection_healthy"):
                try:
                    # For non-RabbitMQ brokers, assume connection is ready after initialization
                    logger.debug("Broker connection assumed to be ready")
                    return
                except Exception as e:
                    logger.debug("Broker connection check failed: %s", e)

            if elapsed_time % 10.0 == 0.0:  # Log every 10 seconds
                logger.warning(
                    "Still waiting for broker connection... (%.1fs elapsed)",
                    elapsed_time,
                )

            await asyncio.sleep(check_interval)
            elapsed_time += check_interval

        raise ConnectionError(
            f"Broker connection not established after {max_wait_time} seconds"
        )
